SystemOrganization addCategory: #MutationEngine!
SystemOrganization addCategory: #'MutationEngine-Tests'!

RBProgramNodeVisitor subclass: #MEConfigurableASTVisitor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEConfigurableASTVisitor commentStamp: 'ag 5/31/2012 11:58:49.547' prior: 0!
This abstract visitor-class uses the abstract method #invokeClient:with: to handle all AST-nodes.
The implementation of that method must return a boolean, indicating, whether or not the client wants to further traverse the tree at that node.!

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:56:40.552'!
acceptArrayNode: node
	
	(self invokeClient: #ArrayNode with: node)
		ifTrue: [ super acceptArrayNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:56:45.386'!
acceptAssignmentNode: node
	
	(self invokeClient: #AssignmentNode with: node)
		ifTrue: [ super acceptAssignmentNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:56:50.46'!
acceptBlockNode: node
	
	(self invokeClient: #BlockNode with: node)
		ifTrue: [ super acceptBlockNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:56:55.655'!
acceptCascadeNode: node
	
	(self invokeClient: #CascadeNode with: node)
		ifTrue: [ super acceptCascadeNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:57:01.543'!
acceptLiteralArrayNode: node
	
	(self invokeClient: #LiteralArrayNode with: node)
		ifTrue: [ super acceptLiteralArrayNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:57:07.023'!
acceptLiteralNode: node
	
	(self invokeClient: #LiteralNode with: node)
		ifTrue: [ super acceptLiteralNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:57:13.966'!
acceptMessageNode: node
	
	(self invokeClient: #MessageNode with: node)
		ifTrue: [ super acceptMessageNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:57:21.516'!
acceptMethodNode: node
	
	(self invokeClient: #MethodNode with: node)
		ifTrue: [ super acceptMethodNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:57:27.525'!
acceptPatternBlockNode: node
	
	(self invokeClient: #PatternBlockNode with: node)
		ifTrue: [ super acceptPatternBlockNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:57:32.897'!
acceptPatternWrapperBlockNode: node
	
	(self invokeClient: #PatternWrapperBlockNode with: node)
		ifTrue: [ super acceptPatternWrapperBlockNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:57:38.686'!
acceptPragmaNode: node
	
	(self invokeClient: #PragmaNode with: node)
		ifTrue: [ super acceptPragmaNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:57:44.319'!
acceptReturnNode: node
	
	(self invokeClient: #ReturnNode with: node)
		ifTrue: [ super acceptReturnNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:57:49.551'!
acceptSequenceNode: node
	
	(self invokeClient: #SequenceNode with: node)
		ifTrue: [ super acceptSequenceNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/31/2012 11:57:55.107'!
acceptVariableNode: node
	
	(self invokeClient: #VariableNode with: node)
		ifTrue: [ super acceptVariableNode: node ].! !

!MEConfigurableASTVisitor methodsFor: 'private' stamp: 'ag 5/31/2012 11:53:43.338'!
clientWantsToVisit: aSymbol
	"Must return a boolean, indicating, whether #doInvokeClient:with: should be called for the given symbol."
	
	self subclassResponsibility.! !

!MEConfigurableASTVisitor methodsFor: 'private' stamp: 'ag 5/31/2012 11:54:23.95'!
doInvokeClient: aSymbol with: aNode
	"If the implementation of this method returns a boolean, it will indicate, whether the AST will be traversed further down
	the given node."
	
	self subclassResponsibility.! !

!MEConfigurableASTVisitor methodsFor: 'private' stamp: 'ag 5/31/2012 11:53:09.592'!
invokeClient: aSymbol with: aNode
	"The return value must be a boolean and indicate, whether we want to continue traversing the tree, or not."
	
	| result |
	result := (self clientWantsToVisit: aSymbol)
		ifTrue: [ self doInvokeClient: aSymbol with: aNode ]
		ifFalse: [ true ].
	(result isKindOf: Boolean) ifFalse: [ result := true ]. "Allow arbitrary results for convenience."
	^ result! !

!MEConfigurableASTVisitor methodsFor: 'visiting' stamp: 'ag 5/31/2012 11:56:22.336'!
visitArgument: node
	"Note -- we invoke accept* on the client, instead of visit*"
	
	(self invokeClient: #Argument with: node)
		ifTrue: [ super visitArgument: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visiting' stamp: 'ag 5/31/2012 11:56:27.434'!
visitArguments: node
	"Note -- we invoke accept* on the client, instead of visit*"
	
	(self invokeClient: #Arguments with: node)
		ifTrue: [ super visitArguments: node ].! !

!MEConfigurableASTVisitor methodsFor: 'visiting' stamp: 'ag 5/31/2012 11:56:32.917'!
visitNode: node
	"Note -- we invoke accept* on the client, instead of visit*"
	
	(self invokeClient: #Node with: node)
		ifTrue: [ super visitNode: node ].! !

MEConfigurableASTVisitor subclass: #MEConfigurableBlockASTVisitor
	instanceVariableNames: 'blocks'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEConfigurableBlockASTVisitor commentStamp: 'ag 7/9/2012 12:41:01.326' prior: 0!
A client using this AST-visitor can configure blocks to be evaluated when nodes of different types are visited.
A simple Dictionary mapping symbols (identifying nodes) to block-objects is maintained internally.
Registered Blocks get up to two arguments (optionally): the visited node and the node-symbol.

The registered symbols are derived from the node-type, e.g. #BlockNode, #IdentifierNode, #Node or #Arguments.
See the instance-side methods of ConfigurableASTVisitor to get a list of these symbols.
!

!MEConfigurableBlockASTVisitor methodsFor: 'private' stamp: 'ag 5/31/2012 11:54:42.967'!
clientWantsToVisit: aSymbol
	
	^ blocks includesKey: aSymbol! !

!MEConfigurableBlockASTVisitor methodsFor: 'private' stamp: 'ag 5/31/2012 11:55:09.399'!
doInvokeClient: aSymbol with: aNode
	
	^ (blocks at: aSymbol)
		valueWithEnoughArguments: { aNode. aSymbol }! !

!MEConfigurableBlockASTVisitor methodsFor: 'initialization' stamp: 'ag 5/25/2012 17:42:24.85'!
initialize
	
	super initialize.
	blocks := Dictionary new.! !

!MEConfigurableBlockASTVisitor methodsFor: 'accessing' stamp: 'ag 5/25/2012 17:47:50.959'!
on: aSymbol invoke: aBlock
	"Configure aBlock to be evaluated, when the visitor visits a node identified by aSymbol.
	aBlock will get up to 2 parameters (the visited node and the symbol identifying it.)"
	
	blocks at: aSymbol put: aBlock.! !

MEConfigurableASTVisitor subclass: #MEConfigurableClientASTVisitor
	instanceVariableNames: 'client'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEConfigurableClientASTVisitor commentStamp: 'ag 5/31/2012 12:00:58.19' prior: 0!
This abstract AST-visitor uses a single client-object to handle the AST-nodes.
For each handled node, a message is sent to that object.
The message name will be: #accept<NODE-TYPE>:
For example: #acceptMessageNode: or acceptNode:
See the instance-side methods of ConfigurableASTVisitor to get a list of possible node-types.!

MEConfigurableClientASTVisitor subclass: #MEConfigurableAutomaticASTVisitor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEConfigurableAutomaticASTVisitor commentStamp: 'ag 7/9/2012 12:41:52.173' prior: 0!
This AST-visitor determines, whether the client-object wants to handle a certain node-type, by checking, whether
it responds to the according #accept* message (see comment in super-class).!

!MEConfigurableAutomaticASTVisitor methodsFor: 'private' stamp: 'ag 5/25/2012 16:39:59.193'!
clientWantsToVisit: aSymbol
	"This implementation uses inspection to tell whether the client wants to visit a certain node."
	
	^ client respondsTo: aSymbol! !

!MEConfigurableClientASTVisitor methodsFor: 'accessing' stamp: 'ag 5/25/2012 17:44:37.297'!
client
	^ client! !

!MEConfigurableClientASTVisitor methodsFor: 'accessing' stamp: 'ag 5/25/2012 17:44:37.58'!
client: anObject
	client := anObject! !

!MEConfigurableClientASTVisitor methodsFor: 'private' stamp: 'ag 5/31/2012 11:58:24.19'!
doInvokeClient: aSymbol with: aNode
	
	^ client perform: ('accept', aSymbol, ':') asSymbol with: aNode! !

MEConfigurableClientASTVisitor subclass: #MEConfigurableSetASTVisitor
	instanceVariableNames: 'interestingNodes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEConfigurableSetASTVisitor commentStamp: 'ag 5/31/2012 12:02:26.278' prior: 0!
This AST-visitor determines, whether the client-object wants to handle a certain node-type, by maintaining a configurable
set of "interesting" node-types, that contains symbols as described in the class-commment of ConfigurableBlockASTVisitor.!

!MEConfigurableSetASTVisitor methodsFor: 'accessing' stamp: 'ag 5/25/2012 16:50:04.784'!
add: aSymbol
	"Add a node-type, that the client is interested in visiting."
	
	interestingNodes add: aSymbol.! !

!MEConfigurableSetASTVisitor methodsFor: 'private' stamp: 'ag 5/25/2012 16:48:44.058'!
clientWantsToVisit: aSymbol
	"This implementation uses a configurable set of symbols, which the client can fill with the node-types, he wants to visit."
	
	^ interestingNodes includes: aSymbol! !

!MEConfigurableSetASTVisitor methodsFor: 'initialization' stamp: 'ag 5/25/2012 16:49:08.67'!
initialize
	
	super initialize.
	interestingNodes := Set new.! !

!MEConfigurableSetASTVisitor methodsFor: 'accessing' stamp: 'ag 5/25/2012 16:50:29.223'!
remove: aSymbol
	"Remove a node-type, that the client is interested in visiting."
	
	interestingNodes remove: aSymbol ifAbsent: [ ^ self ].! !

Object subclass: #MEMutation
	instanceVariableNames: 'originalState parameters injector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEMutation commentStamp: 'ag 7/9/2012 16:07:58.697' prior: 0!
A Mutation object can be applied (#apply) and reverted (#revert).
It can be asked #isApplied, to find out whether it is already applied.
The object can be reused by sending #apply and #revert multiple times.
If #apply is sent multiple times without intermediate #revert, an error-message will propose to revert the Mutation automatically.
#doMutated: can be used to evaluated a Block, while the Mutation is applied, and automatically revert it afterwards.
Important: for each concrete subclass of Mutation, that contains instance variables, #revert should be overwritten to set these variables to an initial state.

Different subclasses of Mutation operate on different parameters. Some of them are required, most of them have default values.
See the class-comments of the subclasses and the methods in the protocols called 'parameters' for further information.

All instance-variables are implementation-details and should not be accessed by the user.
They are initialized by the MutationFactory, that created the Mutation object. See the class-comment of MutationFactory to see, what can be configured.

Following information about the instance-variable for the case, that it's required to create a Mutation without using MutationFactory:
- injector must be an instance of a non-abstract subclass of MutationInjector
- originalState is maintained by the Mutation object itself
- parameters is a dictionary, that must contain at least the key #method. The value is an Association (Class -> Symbol) pointing to the method to mutate.
	Other parameters depend on concrete subclasses of Mutation, see the class-comments and methods in the 'parameters' protocol.
	The class-side protocol 'parameter creation' provides methods to create/test parameter-dictionaries for concrete subclasses.!

MEMutation subclass: #MEAstMutation
	instanceVariableNames: 'ast injectorResult'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEAstMutation commentStamp: 'ag 7/9/2012 16:13:29.636' prior: 0!
This represents Mutations, that explicitely work on the ASTs of methods.
Parts of the abstract methods of Mutation are implemented, and convenience methods, that are expected to be needed by concrete subclasses, are provided.!

!MEAstMutation class methodsFor: 'abstract marking' stamp: 'ag 7/9/2012 12:51:55.699'!
abstract! !

!MEAstMutation methodsFor: 'private' stamp: 'ag 7/9/2012 11:15:24.407'!
applyAst
	
	injectorResult := injector applyAst: self ast toMethod: self method.! !

!MEAstMutation methodsFor: 'accessing' stamp: 'ag 5/31/2012 12:05:00.416'!
ast
	"Return the AST of the mutated method.
	The ast is lazily obtained from the injector-object.
	When the mutation is performed, the modifications are done on this ast-object."
	
	^ ast ifNil: [ ast :=	injector astOfMethod: self method ]! !

!MEAstMutation methodsFor: 'visiting AST nodes' stamp: 'ag 5/25/2012 18:18:15.204'!
nodesToVisit
	"Return a collection of symbols, that identify the types of nodes, that the Mutation-instance is interested
	in visiting. The symbol has to be the Xxx in #acceptXxx: for one of the visitor-methods of
	ConfigurableASTVisitor. A message with the same name will be sent to this Mutation-instance when visiting
	such a node."
	
	self subclassResponsibility.! !

!MEAstMutation methodsFor: 'mutating' stamp: 'ag 7/9/2012 16:21:50.679'!
revert
	
	injectorResult
		ifNotNil: [
			injector revertingAfterHavingApplied: injectorResult.
			injectorResult := nil ].
	ast := nil.
	super revert.! !

!MEAstMutation methodsFor: 'private' stamp: 'ag 7/21/2012 01:24:17.418'!
visitAst
	"Create a shortliving AST-visitor object, using the receiver as client to handle all relevant AST-nodes."
	
	| visitor |
	visitor := MEConfigurableSetASTVisitor new client: self; yourself.
	self nodesToVisit do: [ :symbol | visitor add: symbol asSymbol ].
	visitor visitNode: self ast.! !

MEAstMutation subclass: #MESingleNodeMutation
	instanceVariableNames: 'mutatedNode visitedLiterals'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MESingleNodeMutation commentStamp: 'ag 7/9/2012 16:25:04.533' prior: 0!
This abstract Mutation mutates a single node in the method-AST.
When applying the mutation, the AST is traversed using a Visitor-object, and the mutatedNode is stored (if found).

This class introduces the two parameters nodeSelection and nodeSelectionByIndex.
See the comments in the according methods in the protocol 'parameters' for details.!

MESingleNodeMutation subclass: #MEAddMessageMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEAddMessageMutation commentStamp: 'ag 7/9/2012 16:30:28.498' prior: 0!
An AddMessageMutation takes a ValueNode from the method-AST and appends a unary message to that code-part.
The ValueNode can be almost everything in Smalltalk code (variables, other message sends, literals, ...).
The appended unary message is described by the parameter #selector.

Since ValueNodes can be all kinds of Smalltalk-code-snippets, the #nodeSelection parameter can only contain numbers,
making it difficult to select a certain ValueNode in a certain method by hand.!

MEAddMessageMutation subclass: #MEAddAssignmentMessageMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEAddAssignmentMessageMutation commentStamp: 'ag 7/9/2012 16:35:16.445' prior: 0!
This Mutation adds a unary message-send to the value, that is assigned to a variable.
The message-send is not wrapped around the whole assignment, just around the value-part.

The parameter #nodeSelection can now contain numbers (selecting the nth variable-assignment), and als Strings,
thus selecting the first assignment to the variable (of any kind) with that name.!

!MEAddAssignmentMessageMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:55:32.35'!
nodeValueOf: node
	"The name of the variables, that are assigned to, is used as value for the mutated node.
	The variable name can be used (as String) for the #nodeSelection parameter."
	
	^ node variable name! !

!MEAddAssignmentMessageMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:56:29.098'!
relevantNodeTypes
	"We do not copy the node-types from the superclass!!"
	
	^ #( AssignmentNode )! !

!MEAddAssignmentMessageMutation methodsFor: 'visiting AST nodes' stamp: 'ag 7/5/2012 12:56:16.782'!
acceptAssignmentNode: aNode
	
	^ self doAcceptNode: aNode! !

!MEAddAssignmentMessageMutation methodsFor: 'private' stamp: 'ag 7/5/2012 12:55:53.747'!
mutate: node
	"We do not append the copy-message to the whole assignment node, but only to the value-part."
	
	node value
		replaceWith:
			(super mutate: node value copy).
	^ node! !

MEAddAssignmentMessageMutation subclass: #MEAddCopyAssignmentMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEAddCopyAssignmentMutation commentStamp: 'ag 7/9/2012 17:17:35.955' prior: 0!
This Mutation has the same functionality like the superclass, but the randomly selected selectors,
that are appended to the assignment-values, are copy-selectors (listed in AddCopyAssignmentMutation class >> #copySelectors).

This implements the copy-part of the RAC (reference assignment and copy assignment change) mutation presented in the paper
"A mutation analysis tool for Java programs" by Chevalley and Thevenod-Fosse.!

!MEAddCopyAssignmentMutation class methodsFor: 'constants' stamp: 'ag 7/5/2012 12:55:28.38'!
copySelectors
	
	^ #( copy deepCopy veryDeepCopy )! !

!MEAddCopyAssignmentMutation class methodsFor: 'parameter creation' stamp: 'ag 7/5/2012 12:55:36.522'!
randomParametersFor: methodAST
	
	^ (super randomParametersFor: methodAST),
	{
		#selector -> self copySelectors atRandom
	}! !

!MEAddMessageMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:25:01.647'!
nodeValueOf: aNode
	"Since we take all ValueNodes into account, there is no clear understanding of the value of a node here.
	The parametrer #nodeSelection can only contain a Number."
	
	^ ''! !

!MEAddMessageMutation class methodsFor: 'parameter creation' stamp: 'ag 7/5/2012 12:50:46.419'!
randomParametersFor: methodAST
	
	^ (super randomParametersFor: methodAST),
	{
		"Would be nice to choose a more meaningful selector..."
		#selector -> self randomUnarySelector
	}! !

!MEAddMessageMutation class methodsFor: 'parameter creation' stamp: 'ag 7/5/2012 12:53:41.845'!
randomUnarySelector
	
	| classes |
	classes := Smalltalk allClasses.
	[ classes atRandom selectors
		ifNotEmptyDo: [ :selectors |
			selectors
				detect: [ :symbol | symbol numArgs == 0 ]
				ifFound: [ :symbol | ^ symbol ]
				ifNone: [ "continue..." ] ] ]
	repeat.! !

!MEAddMessageMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:22:50.339'!
relevantNodeTypes
	
	^ #( ValueNode )! !

!MEAddMessageMutation methodsFor: 'visiting AST nodes' stamp: 'ag 7/5/2012 12:29:39.777'!
acceptValueNode: aNode
	
	^ self doAcceptNode: aNode! !

!MEAddMessageMutation methodsFor: 'private' stamp: 'ag 7/5/2012 12:40:09.686'!
mutate: newNode
	"We do not replace the mutated node by newNode directly, but wrap another MessageNode around it."
	
	^ RBMessageNode
		receiver: newNode
		selector: self selector! !

!MEAddMessageMutation methodsFor: 'parameters' stamp: 'ag 7/5/2012 12:41:06.687'!
selector
	"Parameter value: Symbol. Default: #copy.
	This message will be appended to the mutated value-node."
	
	^ self parameters at: #selector ifAbsent: [ #copy ]! !

MESingleNodeMutation subclass: #MERemoveSelectorMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MERemoveSelectorMutation commentStamp: 'ag 7/9/2012 16:43:43.276' prior: 0!
This Mutation removes a whole message-send from the method-AST, including all code-parts forming its parameters.
The parameter #nodeSelection can contain a Symbol, which leads to removal of the first occurrence of that selector in the code.!

MERemoveSelectorMutation subclass: #MERemoveCopySelectorMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MERemoveCopySelectorMutation commentStamp: 'ag 7/9/2012 17:17:30.867' prior: 0!
This Mutation removes a copy-message from the value, that is assigned to a variable. The handled copy-messages are listed in 
AddCopyAssignmentMutation class >> #copySelectors.

This implements the assignment-part of the RAC (reference assignment and copy assignment change) mutation presented in the paper
"A mutation analysis tool for Java programs" by Chevalley and Thevenod-Fosse.!

!MERemoveCopySelectorMutation class methodsFor: 'abstract' stamp: 'ag 7/21/2012 01:26:11.198'!
isRelevantNode: node
	
	^ MEAddCopyAssignmentMutation copySelectors includes: node selector! !

!MERemoveSelectorMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:36:19.634'!
nodeValueOf: node
	"We use the selector as the node-value. This allows to select only certain selectors to be removed."
	
	^ node selector! !

!MERemoveSelectorMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:32:32.893'!
relevantNodeTypes
	
	^ #( MessageNode )! !

!MERemoveSelectorMutation methodsFor: 'visiting AST nodes' stamp: 'ag 7/5/2012 12:34:51.766'!
acceptMessageNode: node
	
	^ self doAcceptNode: node! !

!MERemoveSelectorMutation methodsFor: 'private' stamp: 'ag 7/5/2012 12:59:42.836'!
mutate: node
	"The selected node is a message-node.
	Since we remove the whole message, we simply unpack the receiver node."
	
	^ node receiver! !

!MESingleNodeMutation class methodsFor: 'abstract marking' stamp: 'ag 7/9/2012 12:52:08.621'!
abstract! !

!MESingleNodeMutation class methodsFor: 'support' stamp: 'ag 7/5/2012 12:17:37.487'!
collectLiteralsIn: methodAST
	
	^ (self collectNodesIn: methodAST)
		collect: [ :each | self nodeValueOf: each ]! !

!MESingleNodeMutation class methodsFor: 'support' stamp: 'ag 7/21/2012 01:23:30.545'!
collectNodesIn: methodAST
	
	| visitor nodes |
	nodes := OrderedCollection new.
	visitor := MEConfigurableBlockASTVisitor new.
	self relevantNodeTypes do: [ :symbol |
		visitor
			on: symbol
			invoke: [ :node |
				(self isRelevantNode: node)
					ifTrue: [ nodes add: node ] ] ].
	visitor visitNode: methodAST.
	^ nodes! !

!MESingleNodeMutation class methodsFor: 'constants' stamp: 'ag 7/5/2012 12:13:11.456'!
defaultLiteralSelection
	"By default, select the first appropriate literal."
	
	^ 1! !

!MESingleNodeMutation class methodsFor: 'parameter creation' stamp: 'ag 7/20/2012 14:41:56.224'!
isAbleToMutate: methodAST
	"Check, whether there are any mutatable literals in the given method."
	
	^ (self collectLiteralsIn: methodAST) size > 0! !

!MESingleNodeMutation class methodsFor: 'parameter creation' stamp: 'ag 7/20/2012 14:51:37.508'!
isAbleToMutate: methodAST withParameters: parameters
	"Check, whether there are any mutatable literals in the given method, taking the parameters into account."
	
	| selection literals |
	literals := self collectLiteralsIn: methodAST.
	selection := parameters at: #nodeSelection ifAbsent: [ self defaultLiteralSelection ].
	^ selection isNumber
		ifTrue: [ selection <= literals size ]
		ifFalse: [
			selection isCollection
				ifTrue: [ literals intersects: selection ]
				ifFalse: [ literals includes: selection ] ]! !

!MESingleNodeMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:44:48.996'!
isRelevantNode: anASTNode
	"Answer, whether anASTNode represents a node of the correct type.
	This should define additional constraints besides the node-type, which is already defined over #literalNodeTypes.
	The different subclasses can have different ways of checking this."
	
	^ true! !

!MESingleNodeMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:17:37.331'!
nodeValueOf: aNode
	"Extract the literal value-object from aNode (a sub-instance of RBProgramNode).
	Sadly, subclasses of RBProgramNode do not have a common #value method, that could be used instead of this.
	The subclasses of TokenMutation have different ways of accessing the values in the nodes."
	
	self subclassResponsibility.! !

!MESingleNodeMutation class methodsFor: 'parameter creation' stamp: 'ag 7/5/2012 12:18:12.265'!
randomParametersFor: methodAST
	
	^ (super randomParametersFor: methodAST),
	{
		#nodeSelection -> (self collectLiteralsIn: methodAST) size atRandom
	}! !

!MESingleNodeMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:17:11.158'!
relevantNodeTypes
	"Return the types of the AST-node, that represent the concrete literal value.
	For most cases, this should return a collection with a single element.
	The symbos returned here will be used to parameterize instances of ConfigurableBlockASTVisitor.
	See the class-comment of that class for details about the returned value.
	The appropriate visitor-methods must be implemented by the subclasses."
	
	self subclassResponsibility.! !

!MESingleNodeMutation methodsFor: 'visiting AST nodes' stamp: 'ag 7/5/2012 11:48:33.937'!
acceptNode: node
	"Stop the visiting-process, as soon as the correct literal-node is found.
	If the node is never found, we visit the whole tree."
	
	^ mutatedNode isNil! !

!MESingleNodeMutation methodsFor: 'private' stamp: 'ag 7/5/2012 14:04:47.56'!
applyMutation
	
	| newNode |
	self visitAst.
	mutatedNode ifNil: [ ^ self error: 'The configured AST-node was not found in the mutated method!!' ].
	newNode := self mutate: mutatedNode clone.
	mutatedNode replaceWith: newNode.
	self applyAst.! !

!MESingleNodeMutation methodsFor: 'visiting AST nodes' stamp: 'ag 7/5/2012 12:44:49.175'!
doAcceptNode: node
	"This method is to be called by the according accept-method in the implementing subclasses."
	
	(self class isRelevantNode: node) ifFalse: [ ^ true ]. "Visit only literals of the correct type."
	visitedLiterals := visitedLiterals + 1.
	
	(self isSearchedNode: node)
		ifTrue: [ "Found it!!"
			mutatedNode := node ].
	^ mutatedNode isNil! !

!MESingleNodeMutation methodsFor: 'initialization' stamp: 'ag 7/5/2012 11:47:54.093'!
initialize
	
	super initialize.
	visitedLiterals := 0.! !

!MESingleNodeMutation methodsFor: 'visiting AST nodes' stamp: 'ag 7/5/2012 12:18:28.505'!
isSearchedNode: node
	
	| selection |
	selection := self nodeSelection.
	^ (self nodeSelectionByIndex
		ifTrue: [ visitedLiterals = selection ]
		ifFalse: [ | value |
			value := self class nodeValueOf: node.
			selection isCollection
				ifTrue: [ selection includes: value ]
				ifFalse: [ selection = value ] ])! !

!MESingleNodeMutation methodsFor: 'private' stamp: 'ag 7/5/2012 11:57:23.576'!
mutate: node
	"Perform the required mutation on the node.
	node is already a copy, so modifications can be made directly inside this object.
	The returned object (!!) will be replaced for the original node, so if node should be used itself, it must still be returned explicitely."
	
	self subclassResponsibility.! !

!MESingleNodeMutation methodsFor: 'parameters' stamp: 'ag 7/5/2012 12:18:12.131'!
nodeSelection
	"Parameter type: Number or <literal> or Collection, default: 1 (by default, the first occurrence of the mutated literal-type is selected).
	Defines, how the mutated literal is found in the mutated method.
	A Number value means the n-th literal in that method is selected.
	A literal value means the first literal with that exact value is selected.
	A collection has to include only literal values of the correct type and causes the first included literal, that is found, to be mutated."
	
	^ parameters at: #nodeSelection ifAbsent: [ self class defaultLiteralSelection ]! !

!MESingleNodeMutation methodsFor: 'parameters' stamp: 'ag 7/5/2012 12:18:28.533'!
nodeSelectionByIndex
	"Parameter-type: Boolean, default: <true, depends on #literalSelection>.
	Decides, whether #literalSelection contains a Number value or the actual literal/collection.
	This is mainly needed, to distinguish between these two possibilities in the case of NumberLiteralMutation, since
	#literalSelection will then always contains Number objects."
	
	^ parameters at: #nodeSelectionByIndex ifAbsent: [ self nodeSelection isNumber ]! !

!MESingleNodeMutation methodsFor: 'visiting AST nodes' stamp: 'ag 7/5/2012 12:17:11.263'!
nodesToVisit
	
	^ #( Node ), self class relevantNodeTypes! !

!MESingleNodeMutation methodsFor: 'mutating' stamp: 'ag 7/5/2012 11:48:00.722'!
revert
	
	super revert.
	mutatedNode := nil.
	visitedLiterals := 0.! !

MESingleNodeMutation subclass: #METokenMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!METokenMutation commentStamp: 'ag 7/9/2012 17:30:14.379' prior: 0!
A TokenMutation operates on a single SourceCode-token.
It introduces the parameter #replacementMode, that is evaluated by the implementing subclass and contains Symbol-values, that configure, how the token-value will be mutated. The possible values are described in the class-comments of the concrete subclasses.
!

METokenMutation subclass: #MEIdentifierMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEIdentifierMutation commentStamp: 'ag 7/20/2012 14:44:07.277' prior: 0!
This mutation selects an Identifier from the source-code and replaces it with another one. The replacement-value depends on the #replacementMode parameter, described below.
The value of #nodeSelection can be a String, thus selecting the first occurrence of that identifier in the source-code.
Identifier can be any kind of variable (instance, class), class/global-names, self and super.

Supported replacement modes (see #replacementMode):
	#useSelf
		-- the selected literal is replaced with 'self'
	#useSuper
		-- the selected literal is replaced with 'super'
	#randomInstanceVariable
		-- the selected literal is replaced with a random instance variable from the selected class (including super-classes)
			If the mutated class has no instance-variables at all, a random class will be picked instead.
	#randomClass
		-- the selected literal is replaced with a random class-name
	#replace
		-- the selected literal is replaced with the value of the parameter #modificationValue (see #modificationValue)!

!MEIdentifierMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:17:37.365'!
nodeValueOf: aNode
	
	^ aNode name! !

!MEIdentifierMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:17:11.193'!
relevantNodeTypes
	
	^ #( VariableNode )! !

!MEIdentifierMutation class methodsFor: 'abstract' stamp: 'ag 7/20/2012 14:42:52.574'!
replacementModes
	
	^ #( useSelf useSuper randomClass replace randomInstanceVariable )! !

!MEIdentifierMutation methodsFor: 'visiting AST nodes' stamp: 'ag 7/5/2012 11:54:24.462'!
acceptVariableNode: aNode
	
	^ self doAcceptNode: aNode! !

!MEIdentifierMutation methodsFor: 'private' stamp: 'ag 7/20/2012 15:12:59.038'!
isIllegalMutation: node
	
	^ (node parent isAssignment and: [ node parent variable = node ]
		and: [ (#( useSelf useSuper randomClass ) includes: self replacementMode)
				or: [ #( 'self' 'super' ) includes: (self mutatedValue: 1) ] ])! !

!MEIdentifierMutation methodsFor: 'parameters' stamp: 'ag 6/17/2012 13:55:33.096'!
modificationValue
	"Parameter type: String, default: 'self'.
	When using the replacementMode #replace, the value of this parameter is used to replace the selected
	identifier. This might lead to compilation-errors, if the selected identifier is not a valid global or variable."
	
	^ self parameters at: #modificationValue ifAbsent: [ 'self' ]! !

!MEIdentifierMutation methodsFor: 'private' stamp: 'ag 7/20/2012 15:13:26.948'!
mutate: node
	"Include an additional sanity-check here. If we are mutating the left side of an assignment, 
	we cannot use self/super or any class-name as the mutated value (cannot store into these identifiers).
	At this place, we cannot find an alternative solution anymore, so we just skip the mutation. Sorry."
	
	^ (self isIllegalMutation: node)
		ifTrue: [ node ]
		ifFalse: [ super mutate: node ]! !

!MEIdentifierMutation methodsFor: 'private' stamp: 'ag 7/20/2012 14:55:13.454'!
mutatedValue: identifier
	
	^ self replacementMode
		caseOf: {
			[ #useSelf ] -> [ 'self' ].
			[ #useSuper ] -> [ 'super' ].
			[ #randomInstanceVariable ] -> [
				self method key instVarNames
					ifEmpty: [ self randomClassName ] "Do not fail, if the current class has no instance-variables; fall back to a random class instead."
					ifNotEmptyDo: #atRandom ].
			[ #randomClass ] -> [ self randomClassName ].
			[ #replace ] -> [ self modificationValue ].
		}! !

!MEIdentifierMutation methodsFor: 'private' stamp: 'ag 7/20/2012 14:43:44.005'!
randomClassName
	
	^ Smalltalk allClasses atRandom name asString! !

MEIdentifierMutation subclass: #MESelfSuperMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MESelfSuperMutation commentStamp: 'ag 7/9/2012 17:38:36.481' prior: 0!
This Mutation selects only occurrences of self and super for mutation.
The rest of the behavior is the same as the super-class.

This implements the ide (identifier replacement) mutation presented in the paper
"Class Mutation: Mutation Testing for Object-Oriented Programs" by Kim, Clark and McDermid.

This also covers the ice (instance creation mutation) mutation from the same paper,
since instances are created by sending messages to classes, and this Mutation can change the class-name of such an expression.!

!MESelfSuperMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:44:49.031'!
isRelevantNode: anASTNode
	"This specialization of the IdentifierMutation only deals with the self and super keywords."
	
	^ #( 'self' 'super' ) includes: anASTNode name! !

METokenMutation subclass: #MEKeywordArgumentMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEKeywordArgumentMutation commentStamp: 'ag 7/9/2012 17:40:22.564' prior: 0!
The KeywordArgumentMutation changes the arguments of a keyword message.
The value of the parameter #nodeSelection can be a Symbol; the first occurrence of that symbol is then selected for mutation.

!

MEKeywordArgumentMutation subclass: #MEArgumentNumberDecreaseMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEArgumentNumberDecreaseMutation commentStamp: 'ag 7/9/2012 17:46:46.442' prior: 0!
This Mutation deletes a part of the arguments of a keyword message-send.
The deletion of one or more arguments also leads to a change of the selector. This might destroy the program semantics.
The removed arguments are described by the parameters #start and #decreaseBy, which give the start-index in the arguments and the number of arguments to delete starting from that index. This means, that arguments can be deleted from the middle of the selector.
If enough arguments are deleted, the keyword-message might become a unary-message.

The single #replacementMode (#decrease) in this class is not functional and only defined to support the mechanisms defined by the super-classes.!

!MEArgumentNumberDecreaseMutation class methodsFor: 'abstract' stamp: 'ag 7/2/2012 10:37:48.401'!
defaultReplacementMode
	"(Irrelevant value, replacement mode not evaluated)"
	
	^ #decrease! !

!MEArgumentNumberDecreaseMutation class methodsFor: 'parameter creation' stamp: 'ag 7/5/2012 12:18:12.291'!
randomParametersFor: methodAST
	
	| parameters messageNode start |
	parameters := super randomParametersFor: methodAST.
	messageNode := (self collectNodesIn: methodAST) at: (parameters at: #nodeSelection).
	start := messageNode arguments size atRandom.
	^ parameters,
	{
		#start -> start.
		#decreaseBy -> (messageNode arguments size - start) atRandom
	}! !

!MEArgumentNumberDecreaseMutation class methodsFor: 'abstract' stamp: 'ag 7/2/2012 10:37:52.714'!
replacementModes
	"(Irrelevant value, replacement mode not evaluated)"
	
	^ #( decrease )! !

!MEArgumentNumberDecreaseMutation methodsFor: 'private' stamp: 'ag 6/24/2012 19:22:36.622'!
asUnarySelector: symbol
	
	^ symbol copyFrom: 1 to: (symbol indexOf: $:) - 1! !

!MEArgumentNumberDecreaseMutation methodsFor: 'parameters' stamp: 'ag 6/24/2012 18:51:07.25'!
decreaseBy
	"Parameter value: Number. Default: 1.
	The number of arguments, that will be removed from the mutated keyword-message-send.
	This will later be ensured to be at least one, and at most the number of arguments the mutations message has."
	
	^ self parameters at: #decreaseBy ifAbsent: [ 1 ]! !

!MEArgumentNumberDecreaseMutation methodsFor: 'private' stamp: 'ag 7/2/2012 12:58:15.311'!
mutatedValue: originalValue
	"The return value is not a symbol, like the parameter, but an Association from the symbol to a collection of argument-nodes,
	that represent the mutated arguments of the message-send, in the correct order."
	
	| newSelector newArguments start end numArguments selectorParts |
	numArguments := originalValue count: [ :char | char = $: ].
	start := self start clampLow: 1 high: numArguments.
	end := (start + self decreaseBy) clampHigh: numArguments.
	newArguments := OrderedCollection new.
	newSelector := ''.
	selectorParts := self splitSelector: originalValue.
	1 to: numArguments do: [ :i |
		(i between: start and: end)
			ifFalse: [
				newSelector := newSelector, (selectorParts at: i).
				newArguments add: (mutatedNode arguments at: i) ] ].
	newArguments ifEmpty: [ "No arguments left. Take the first part of the selector, making it a unary message."
		newSelector := self asUnarySelector: originalValue ].
	^ newSelector asSymbol -> newArguments! !

!MEArgumentNumberDecreaseMutation methodsFor: 'private' stamp: 'ag 7/2/2012 12:56:35.329'!
selectorPart: index of: symbol
	
	! !

!MEArgumentNumberDecreaseMutation methodsFor: 'parameters' stamp: 'ag 6/24/2012 19:30:25.534'!
start
	"Parameter value: Number. Default: 1.
	The parameter-number, which will be the first to be deleted. Starting from this parameter,
	#decreaseBy parameters will be deleted. Parameter indexing is one-based, like collections.
	The selection of parameters to delete is NOT swapped around the border of the interval [0..numArgs].
	E.g., if you start at the last argument and #decreaseBy is set to 10, still only the last argument will be deleted."
	
	^ self parameters at: #start ifAbsent: [ 1 ]! !

MEKeywordArgumentMutation subclass: #MEArgumentNumberIncreaseMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEArgumentNumberIncreaseMutation commentStamp: 'ag 7/9/2012 17:49:41.27' prior: 0!
This Mutation adds new parameters to a keyword-message.
In addition to the argument-values, the new selector-parts can be given as well.
The configuration of this Mutation is rather complicated and is described in the single parameter #additionalArguments.

The single replacementMode (#increase) in this class is not functional and only defined to support the mechanisms defined by the super-classes.!

!MEArgumentNumberIncreaseMutation class methodsFor: 'abstract' stamp: 'ag 7/2/2012 10:39:10.521'!
defaultReplacementMode
	"(Irrelevant value, replacement mode not evaluated)"
	
	^ #increase! !

!MEArgumentNumberIncreaseMutation class methodsFor: 'parameter creation' stamp: 'ag 7/21/2012 01:28:39.527'!
randomLiteralValue
	
	^ #(string number keyword) atRandom
		caseOf: {
			[ #string ] -> [ MEStringLiteralMutation randomString: 6 atRandom ].
			[ #number ] -> [ MENumberLiteralMutation randomNumber ].
			[ #keyword ] -> [ #(true false nil) atRandom ]
		}! !

!MEArgumentNumberIncreaseMutation class methodsFor: 'parameter creation' stamp: 'ag 7/5/2012 12:18:12.102'!
randomParametersFor: methodAST
	
	| parameters messageNode start args |
	parameters := super randomParametersFor: methodAST.
	messageNode := (self collectNodesIn: methodAST) at: (parameters at: #nodeSelection).
	start := messageNode arguments size atRandom.
	args := OrderedCollection new.
	3 atRandom timesRepeat: [
		args add: {
			messageNode arguments size atRandom. "Random index for new parameter"
			#('aa' 'bb' 'cc' 'dd' 'ee') atRandom. "Pseudo-random additional selector-part"
			self randomLiteralValue } ]. "Pseudo-random literal value for the argument"
	^ parameters,
	{
		#additionalArguments -> args
	}! !

!MEArgumentNumberIncreaseMutation class methodsFor: 'abstract' stamp: 'ag 7/2/2012 10:39:06.328'!
replacementModes
	"(Irrelevant value, replacement mode not evaluated)"
	
	^ #( increase )! !

!MEArgumentNumberIncreaseMutation methodsFor: 'parameters' stamp: 'ag 7/2/2012 12:15:52.676'!
additionalArguments
	"Parameter value: Collection of <Array<size 3: Number, Symbol, String/Number/nil/Boolean>. Default: empty.
	The 3-tuples in this collection describe the additional arguments, that will be inserted into the 
	mutated keyword-message. Every item stands for one additional argument.
	The 3-tuples contain the following:
	 - The first element gives the index, where to insert the new argument. (< 0) -> prepend it, (> original number of arguments) (or nil) -> append it.
	 - The second element gives the selector-part to be used for this argument. A colon will be appended, if not present.
	 - The third element gives the parameter-value. It must be an object convertable to a simple literal AST-node."
	
	^ self parameters at: #additionalArguments ifAbsentPut: [ OrderedCollection new ]! !

!MEArgumentNumberIncreaseMutation methodsFor: 'private' stamp: 'ag 7/9/2012 17:52:32.976'!
buildNodeFrom: object
	
	^ object isNumber
		ifTrue: [ RBLiteralValueNode literalToken: (RBNumberLiteralToken value: object) ]
		ifFalse: [ RBLiteralValueNode value: object ]! !

!MEArgumentNumberIncreaseMutation methodsFor: 'private' stamp: 'ag 7/2/2012 14:39:13.582'!
mutatedValue: originalValue
	
	| newSelector newArguments originalNumArguments |
	newSelector := self splitSelector: originalValue.
	newArguments := mutatedNode arguments copy.
	originalNumArguments := newArguments size.
	self additionalArguments do: [ :arg | | index |
		index := arg first.
		index < 0 ifTrue: [ index := 1 ].
		index > originalNumArguments ifTrue: [ index := newArguments size + 1 ].
		newArguments add: (self buildNodeFrom: arg third) beforeIndex: index.
		newSelector add: arg second beforeIndex: index ].
	^ (self buildSelector: newSelector) -> newArguments! !

MEKeywordArgumentMutation subclass: #MEArgumentOrderMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEArgumentOrderMutation commentStamp: 'ag 7/9/2012 17:54:20.864' prior: 0!
This mutation changes the order of arguments in a keyword-message-send.
The way the arguments are reordered is mainly dependant of the #replacementMode, see below.

Supported replacement modes (see #replacementMode):
	#rotateRight
		-- The arguments of the mutated keyword-message are rotated to the right (by the value of #rotateBy)
	#rotateLeft
		-- The arguments of the mutated keyword-message are rotated to the left (by the value of #rotateBy)
	#shuffle
		-- The argument-order of the mutated keyword-message is determined by shuffling the arguments.!

!MEArgumentOrderMutation class methodsFor: 'abstract' stamp: 'ag 6/23/2012 17:39:13.243'!
defaultReplacementMode
	
	^ #shuffle! !

!MEArgumentOrderMutation class methodsFor: 'parameter creation' stamp: 'ag 7/5/2012 12:18:12.074'!
randomParametersFor: methodAST
	
	| parameters messageNode |
	parameters := super randomParametersFor: methodAST.
	messageNode := (self collectNodesIn: methodAST) at: (parameters at: #nodeSelection).
	^ parameters,
	{
		#rotateBy -> messageNode arguments size atRandom
	}! !

!MEArgumentOrderMutation class methodsFor: 'abstract' stamp: 'ag 7/2/2012 10:38:48.704'!
replacementModes
	
	^ #( rotateRight rotateLeft shuffle )! !

!MEArgumentOrderMutation methodsFor: 'private' stamp: 'ag 7/2/2012 12:07:32.507'!
mutatedValue: originalValue
	
	| arguments |
	arguments := mutatedNode arguments.
	^ originalValue -> 
		(self replacementMode
			caseOf: {
				[ #shuffle ] -> [ arguments copy shuffle; yourself ].
				[ #rotateLeft ] -> [ self rotate: arguments by: self rotateBy negated ].
				[ #rotateRight ] -> [ self rotate: arguments by: self rotateBy ].
			})! !

!MEArgumentOrderMutation methodsFor: 'private' stamp: 'ag 7/2/2012 12:06:36.773'!
rotate: collection by: number
	
	| result |
	result := collection species ofSize: collection size.
	collection withIndexDo: [ :each :i |
		result atWrap: i + number put: each ].
	^ result! !

!MEArgumentOrderMutation methodsFor: 'parameters' stamp: 'ag 7/2/2012 10:44:58.857'!
rotateBy
	"Parameter value: Number, default: 1.
	If using #replacementMode rotateRight or rotateLeft, this parameter gives the number of times, that the parameters will be
	shifted in the according direction."
	
	^ self parameters at: #rotateBy ifAbsent: [ 1 ]! !

!MEKeywordArgumentMutation class methodsFor: 'abstract marking' stamp: 'ag 7/9/2012 12:52:53.662'!
abstract! !

!MEKeywordArgumentMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:44:48.913'!
isRelevantNode: anASTNode
	"Not every message-send is handled, only keyword-messages.
	Unary message could be handled here as well, but the only #replacement mode applicable 
	for those would be #add (because they don't have any arguments to #shuffle/#delete/etc.)"
	
	^ anASTNode isKeyword! !

!MEKeywordArgumentMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:17:37.257'!
nodeValueOf: aNode
	
	^ aNode selector! !

!MEKeywordArgumentMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:17:11.087'!
relevantNodeTypes
	
	^ #( MessageNode )! !

!MEKeywordArgumentMutation methodsFor: 'visiting AST nodes' stamp: 'ag 7/5/2012 11:54:24.488'!
acceptMessageNode: aNode
	
	^ self doAcceptNode: aNode! !

!MEKeywordArgumentMutation methodsFor: 'private' stamp: 'ag 7/2/2012 12:59:56.812'!
buildSelector: collection
	
	^ (collection
		inject: ''
		into: [ :all :each | all, each, (each last = $: ifTrue: [ '' ] ifFalse: [ ':' ]) ])
			asSymbol! !

!MEKeywordArgumentMutation methodsFor: 'private' stamp: 'ag 7/2/2012 10:41:01.11'!
changeValueOf: newNode to: newValue
	
	newNode
		renameSelector: newValue key
		andArguments: newValue value.! !

!MEKeywordArgumentMutation methodsFor: 'private' stamp: 'ag 7/2/2012 12:57:18.472'!
splitSelector: aSymbol
	
	^ (aSymbol findTokens: ':') collect: [ :part | part, ':' ]! !

METokenMutation subclass: #MELiteralMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MELiteralMutation commentStamp: 'ag 7/9/2012 17:54:43.169' prior: 0!
A LiteralMutation operates on a single source-code literal like a String or a Number.!

MELiteralMutation subclass: #MEKeywordMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEKeywordMutation commentStamp: 'ag 7/9/2012 17:58:46.92' prior: 0!
A KeywordMutation mutates a Smalltalk-language keyword.
Smalltalk does not have too many keywords; the ones that are handled here are nil, true and false.
self and super are handled by the IdentifierMutation. The reasen, that they are not handled here is mainly because they are represented
as instances of IdentifierNode by the AST library.
The vale of the #nodeSelection parameter can be one of nil/true/false; the first occurrence of that literal will be mutated.

Supported replacement modes (see #replacementMode):
	#random
		-- replaces the keyword with a random value from { nil. true. false }.
	#preserveType
		-- replaces the keyword with another one of the same type: (true <-> false). nil stays nil.!

!MEKeywordMutation class methodsFor: 'abstract' stamp: 'ag 5/31/2012 16:28:20.278'!
defaultReplacementMode
	
	^ #preserveType! !

!MEKeywordMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:44:49.138'!
isRelevantNode: aNode
	"This mutation handles part of the keywords in the Smalltalk-language, namely those, that are represented through a literal-node.
	self/super are represented as variable-nodes, so they can be mutated using the IdentifierMutation."
	
	^ aNode isLiteral and: [ #( true false nil ) includes: aNode value ]! !

!MEKeywordMutation class methodsFor: 'abstract' stamp: 'ag 5/31/2012 16:30:41.51'!
replacementModes
	
	^ #( random preserveType )! !

!MEKeywordMutation methodsFor: 'private' stamp: 'ag 7/9/2012 11:42:36.621'!
mutatedValue: anObject
	"The mutatedValue is held as an instance variable, because this message might get called twice and then has to produce the same result.
	This comes from the kind of awkward fact, that this Mutation operates on Literal, as well as on Identifier nodes, to cover all of true,false,nil,self,super."
	
	^ self replacementMode
		caseOf: {
			[ #random ] -> [ #( true false nil ) asSet remove: anObject; atRandom ].
			[ #preserveType ] -> [
				(#( true false ) includes: anObject)
					ifTrue: [ anObject not ]
					ifFalse: [ anObject "This only leaves nil, which is preserved." ] ] }! !

!MELiteralMutation class methodsFor: 'abstract marking' stamp: 'ag 7/9/2012 12:52:43.624'!
abstract! !

!MELiteralMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:17:37.399'!
nodeValueOf: aNode  
	
	^ aNode value! !

!MELiteralMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:17:11.228'!
relevantNodeTypes
	
	^ #( LiteralNode )! !

!MELiteralMutation methodsFor: 'visiting AST nodes' stamp: 'ag 7/5/2012 11:54:24.514'!
acceptLiteralNode: node
	
	^ self doAcceptNode: node! !

!MELiteralMutation methodsFor: 'private' stamp: 'ag 7/5/2012 14:02:48.089'!
changeValueOf: newNode to: newValue
	
	super changeValueOf: newNode to: newValue.
	newNode token
		stop: newNode token start + newValue asString size.! !

MELiteralMutation subclass: #MENumberLiteralMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MENumberLiteralMutation commentStamp: 'ag 7/9/2012 17:57:52.677' prior: 0!
This Mutation changes a Number literal to another Number.
The parameter #nodeSelectionByIndex is important here - since #nodeSelection is always a number,
the user has to specify, whether he is selecting a certain Number literal, or the nth Number literal.

Supported replacement modes (see #replacementMode):
	#random
		-- replaces the number literal with a random integer value (NumberLiteralMutation >> #randomNumber)
	#smaller
		-- replaces the number literal with a random number, that is smaller than the original number
	#bigger
		-- replaces the number literal with a random number, that is bigger than the original number
	#increment
		-- adds the value of the parameter #modificationValue to the original number-literal.
	#decrement
		-- subtracts the value of the parameter #modificationValue from the original number-literal.
	instance of Number -> use this parameter value directly as the mutated value.
	
(for details about #modificationValue, see NumberLiteralMutation >> #modificationValue)!

!MENumberLiteralMutation class methodsFor: 'abstract' stamp: 'ag 5/31/2012 14:03:57.199'!
defaultReplacementMode
	"Default replacement mode for number-literals. See #replacementMode in LiteralMutation, or class-comment of NumberLiteralMutation."
	
	^ #random! !

!MENumberLiteralMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:44:48.96'!
isRelevantNode: anASTNode
	
	^ anASTNode value isNumber! !

!MENumberLiteralMutation class methodsFor: 'parameter creation' stamp: 'ag 5/31/2012 15:50:55.685'!
randomNumber
	"The literal here is the biggest number, found by manual testing, that evaluates to SmallInteger."
	
	^ 16r3fffffff atRandom! !

!MENumberLiteralMutation class methodsFor: 'parameter creation' stamp: 'ag 6/17/2012 11:50:11.031'!
randomParametersFor: methodAST
	
	^ (super randomParametersFor: methodAST),
	{
		#modificationValue -> (self randomNumber / 2)
	}! !

!MENumberLiteralMutation class methodsFor: 'abstract' stamp: 'ag 5/31/2012 16:25:13.879'!
replacementModes
	
	^ #( random increment decrement smaller bigger )! !

!MENumberLiteralMutation methodsFor: 'private' stamp: 'ag 5/31/2012 14:51:36.842'!
changeValueOf: newNode to: newValue
	
	super changeValueOf: newNode to: newValue.
	newNode token source: newValue asString.! !

!MENumberLiteralMutation methodsFor: 'parameters' stamp: 'ag 5/31/2012 14:17:26.022'!
modificationValue
	"Parameter type: Number, default: 1.
	This parameter configures, what is added/subtracted from the original value when using 
	#increment or #decrement as #replacementMode."
	
	^ parameters at: #modificationValue ifAbsent: [ 1 ]! !

!MENumberLiteralMutation methodsFor: 'private' stamp: 'ag 7/5/2012 14:30:56.405'!
mutatedValue: aNumber
	
	self replacementMode isNumber
		ifTrue: [ ^ self replacementMode ]
		ifFalse: [
			^ self replacementMode
				caseOf: {
					[ #random ] -> [ self class randomNumber ].
					[ #increment ] -> [ aNumber + self modificationValue ].
					[ #decrement ] -> [ aNumber - self modificationValue ].
					[ #bigger ] -> [ aNumber + self class randomNumber ].
					[ #smaller ] -> [ aNumber - self class randomNumber ].
				} ]! !

MELiteralMutation subclass: #MEStringLiteralMutation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEStringLiteralMutation commentStamp: 'ag 7/9/2012 18:03:20.601' prior: 0!
This Mutation replaces a String literal with another value.
The value of the #nodeSelection parameter can be a String object; the first occurrence of that String will then be mutated.

Supported replacement modes (see #replacementMode):
	Random characters are printable ASCII characters (range 33..126, see #randomCharacter).
	#shuffle
			-- The whole string is shuffled, preserving the string length.
	#randomString
			-- A random length is selected (maximum: original length * 2), and populated with random characters.
	#randomlyReplaceCharacters
			-- Every character (except white space) in the mutated String-literal is replaced by a random other character (except white space).

If the value of #replacementMode is an instance of String (not Symbol, #isSymbol = false) -> use this parameter value directly as the mutated value.!

!MEStringLiteralMutation class methodsFor: 'random parameter creation' stamp: 'ag 7/5/2012 14:25:16.783'!
change: aString to: aBlock
	"Create some random String using aBlock and make sure, that it is not equal to the original String."
	
	| result |
	result := aBlock value.
	[ result = aString ] whileTrue: [ result := aBlock value ].
	^ result! !

!MEStringLiteralMutation class methodsFor: 'abstract' stamp: 'ag 5/31/2012 13:58:46.42'!
defaultReplacementMode
	"Default replacement mode for string-literals. See #replacementMode in LiteralMutation, or class-comment of StringLiteralMutation."
	
	^ #shuffle! !

!MEStringLiteralMutation class methodsFor: 'abstract' stamp: 'ag 7/5/2012 12:44:49.067'!
isRelevantNode: anASTNode
	
	^ anASTNode value isString! !

!MEStringLiteralMutation class methodsFor: 'random parameter creation' stamp: 'ag 7/2/2012 14:35:09.169'!
randomCharacter
	"Pick a random character, choose only printable characters.
	The ASCII table starts with 32 non-printables, and ends with one non-printable.
	Also, atRandom does not return zero as value."
	
	^ ((128 - 33) atRandom + 31) asCharacter! !

!MEStringLiteralMutation class methodsFor: 'random parameter creation' stamp: 'ag 7/2/2012 14:34:26.536'!
randomString: length
	
	| result |
	result := String new: length.
	1 to: length do: [ :i | 
		result at: i put: self randomCharacter ].
	^ result! !

!MEStringLiteralMutation class methodsFor: 'abstract' stamp: 'ag 5/31/2012 16:25:23.604'!
replacementModes
	
	^ #( shuffle randomString randomlyReplaceCharacters )! !

!MEStringLiteralMutation methodsFor: 'private' stamp: 'ag 7/5/2012 14:26:20.57'!
mutatedValue: aString
	"Yeah I know... #caseOf:... But it's really what we need here!!"
	
	^ self replacementMode isSymbol
		ifFalse: [ ^ self replacementMode asString ]
		ifTrue: [
			self replacementMode
				caseOf: {
					[#randomlyReplaceCharacters] -> [
						String streamContents: [ :stream |
							aString do: [ :char |
								char isSeparator
									ifTrue: [ stream nextPut: char ]
									ifFalse: [ stream nextPut: self class randomCharacter ] ] ] ].
					[#randomString] -> [
						self class
							change: aString
							to: [ self class randomString: (aString size * 2) atRandom ] ].
					[#shuffle] -> [
						self class
							change: aString
							to: [ aString shuffled ] ]
				} ]! !

!METokenMutation class methodsFor: 'abstract marking' stamp: 'ag 7/9/2012 12:52:26.972'!
abstract! !

!METokenMutation class methodsFor: 'abstract' stamp: 'ag 5/31/2012 13:32:00.069'!
defaultReplacementMode
	
	self subclassResponsibility.! !

!METokenMutation class methodsFor: 'parameter creation' stamp: 'ag 7/5/2012 12:13:34.318'!
randomParametersFor: methodAST
	
	^ (super randomParametersFor: methodAST),
	{
		#replacementMode -> self replacementModes atRandom.
	}! !

!METokenMutation class methodsFor: 'abstract' stamp: 'ag 5/31/2012 13:57:04.278'!
replacementModes
	"See instance-side, #replacementMode"
	
	self subclassResponsibility.! !

!METokenMutation methodsFor: 'private' stamp: 'ag 6/17/2012 14:01:53.859'!
changeValueOf: newNode to: newValue
	"Set the newValue object as the literal value of newNode."
	
	| token |
	token := newNode token.
	token
		value: newValue
		start: token start.! !

!METokenMutation methodsFor: 'private' stamp: 'ag 7/5/2012 14:10:37.886'!
mutate: node
	
	self
		changeValueOf: node
		to: (self mutatedValue: (self class nodeValueOf: node)).
	^ node! !

!METokenMutation methodsFor: 'private' stamp: 'ag 5/31/2012 13:47:26.121'!
mutatedValue: originalValue
	"Answer a an object, that should replace the original literal-value.
	The class of the originalValue object depends on the concrete subclass."
	
	self subclassResponsibility.! !

!METokenMutation methodsFor: 'parameters' stamp: 'ag 7/5/2012 14:30:05.891'!
replacementMode
	"Parameter type: Symbol, default depends on concrete subclass.
	See the class-comments of the subclasses.
	Defines, how the literal is mutated.
	Alternatively, most subclasses allow an Object of the mutated type, to explicitely define the
	mutation result (e.g. a Number object for a NumberLiteralMutation)."
	
	^ parameters at: #replacementMode ifAbsent: [ self class defaultReplacementMode ]! !

!MEMutation class methodsFor: 'abstract marking' stamp: 'ag 7/9/2012 12:51:26.974'!
abstract
	"This method can be added in subclasses, that are abstract and cannot function by their own.
	Nothing is needed to be returned here, the method-dictionaries are just checked to include this selector."! !

!MEMutation class methodsFor: 'instance creation' stamp: 'ag 5/25/2012 14:42:01.678'!
injector: aMutationInjector parameters: aDictionary
	
	^ self basicNew
		injector: aMutationInjector;
		parameters: aDictionary;
		initialize; "Initialize after setting the parameters-dictionary."
		yourself! !

!MEMutation class methodsFor: 'parameter creation' stamp: 'ag 7/9/2012 16:02:12.492'!
isAbleToMutate: aMethod
	"Return a boolean value, indicating whether this mutation-class would be able to mutate the given method.
	The parameters for the mutation will be selected randomly (#randomParametersFor:), so just the minimum requirements
	for the mutations have to be checked.
	Example: for a StringLiteralMutation, check whether aMethod contains ANY String-literal."
	
	self subclassResponsibility.! !

!MEMutation class methodsFor: 'parameter creation' stamp: 'ag 7/9/2012 16:03:43.194'!
isAbleToMutate: aMethod withParameters: parameters
	"Return a boolean value, indicating whether this mutation-class would be able to mutate the given method using the given parameters.
	Example: For StringLiteralMutation, check, whether aMethod contains a String-literal, as described by the #nodeSelection parameter in the parameters-dictionary."
	
	self subclassResponsibility.! !

!MEMutation class methodsFor: 'abstract marking' stamp: 'ag 7/9/2012 15:47:49.06'!
isAbstract
	
	^ self class methodDict includesKey: #abstract! !

!MEMutation class methodsFor: 'parameter creation' stamp: 'ag 7/9/2012 16:06:17.335'!
randomParametersFor: methodAST
	"Subclasses should implement this, and return a Dictionary with random parameters, that are meaningful for 
	instances of those mutation-classes. This is used by a MutationFactory, when creating random mutation-objects.
	All parameters, that are required by the mutation, have to be created. Optional parameters can be left out.
	
	If the subclass decides, that the Mutation cannot be applied on the given method, then nil must be returned.
	The parameters can be dependant on the mutated method, that is given as AST.
	
	Following statement should always be true for subclasses X of Mutation:
	(X isAbleToMutate: methodAST)
	==
	(X isAbleToMutate: methodAST withParameters: (X randomParametersFor: methodAST))"
	
	^ Dictionary new! !

!MEMutation methodsFor: 'mutating' stamp: 'ag 7/5/2012 13:09:41.037'!
apply
	"Store the current state of the mutated method and apply the mutation.
	#revert has to be called afterwards, to restore the original state.
	To make sure, that the state is really reverted, exceptions between #apply and #revert should be handled carefullly.
	#doMutated: can be used for that."
	
	self isApplied ifTrue:
		[ self error: 
				'This mutation-object has already been applied!! ',
				'It is unsafe to apply it again, #revert must be called first. ',
				'Proceed to revert this mutation and re-apply it automatically.'.
		self revert ].
	originalState := injector currentStateOfMethod: self method.
	self applyMutation.! !

!MEMutation methodsFor: 'private' stamp: 'ag 5/25/2012 18:18:52.972'!
applyMutation
	"This is the hook for subclasses to actually perform the mutation on the AST, obtained from the injector variable."
	
	self subclassResponsibility.! !

!MEMutation methodsFor: 'mutating' stamp: 'ag 5/25/2012 17:28:08.127'!
doMutated: aBlock
	"Evaluated aBlock, with the mutation active. Any exceptions from aBlock are caught to
	make sure, that the mutation is reverted afterwards."
	
	self apply.
	^ aBlock ensure: [ self revert ]! !

!MEMutation methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:14:56.444'!
injector
	^ injector! !

!MEMutation methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:14:56.475'!
injector: anObject
	injector := anObject! !

!MEMutation methodsFor: 'accessing' stamp: 'ag 7/5/2012 13:09:51.387'!
isApplied
	
	^ originalState notNil! !

!MEMutation methodsFor: 'parameters' stamp: 'ag 5/25/2012 14:25:45.021'!
method
	"This parameter indicates the method to be mutated.
	Value: Association (Class -> selector-symbol).
	This parameter is required for all mutations, that change source-code."
	
	^ parameters at: #method! !

!MEMutation methodsFor: 'accessing' stamp: 'ag 7/5/2012 12:47:02.186'!
mutatedClass
	
	^ self method key! !

!MEMutation methodsFor: 'accessing' stamp: 'ag 7/5/2012 12:47:09.204'!
mutatedSelector
	
	^ self method value! !

!MEMutation methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:14:56.387'!
parameters
	^ parameters! !

!MEMutation methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:14:56.414'!
parameters: anObject
	parameters := anObject! !

!MEMutation methodsFor: 'mutating' stamp: 'ag 5/25/2012 14:42:34.38'!
revert
	"Bring the mutated method to its original state. Only shows any effect, if #apply has been called earlier."
	
	originalState ifNotNil: [
		injector restoreStateOfMethod: self method to: originalState.
		originalState := nil "Mark this mutation-object as reverted, to prevent restoring the state multiple times." ].! !

Object subclass: #MEMutationFactory
	instanceVariableNames: 'injector scope availableMutations'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEMutationFactory commentStamp: 'ag 7/9/2012 12:22:28.231' prior: 0!
An instance MutationFactory can produce Mutation objects.
The methods for this ar eplaced in the protocol 'mutation creation'. A MutationFactory is created using the methods in the 'instance creation' protocol on the class side. See below (paragraph about injector) for the meaning of the different methods there.
Different parameters can be specified by the user or selected randomly by the factory.
Whenever a parameter is selected randomly, care is taken, that the resulting Mutation object will actually be able to perform the mutation.
When the given parameters do not allow creation of a working Mutation (e.g. when trying to create a StringLiteralMutation on a method, that does not contain any string literals), nil will be returned by the factory-method.
The methods to create Mutation objects are located in the 'mutation creation' protocol.
The following aspects can be configured in the MutationFactory.

== The Mutation class (availableMutations)
A Mutation factory has a collection of subclasses of Mutation (availableMutations), from which it will select randomly, if a random Mutations is required.
If the mutation class is given by the user to the factory-method, this collection is not used.

== The Method scope (scope)
A Mutation factory has an instance of MutationScope (scope). This object can be configured by itself (see the comment in MutationScope).
Whenever a mutation for a random location (method) is queried, the scope is queried for a random method. The scope is then required to be set by the user.
If the mutation-location is given by the user, this object is not used.

== injector
This instance of MutationInjector is passed directly to every created Mutation.
This is an internal implementation object, separating certain concerns from the Mutation class. See the class comment in MutationInjector.
Which implementation of the MutationInjector is used by the factory is determined by the instance-creation-method used to obtain the MutationFactory object:
	onSourcecode -> SourcecodeMutationInjector
	onMethodWrappers -> MethodWrapperMutationInjecwithInjector
	withInjector: -> any MutationInjector implementation
	
== The Mutation parameters
The parameters for a Mutation object can be passed to the factory-methods of the MutationFactory (protocol 'mutation creation').
If a method without parameters is used, random appropriate parameters will be generated by the Mutation class, before creating the Mutation object.
In that case, care will be taken, that the parameters wil actually produce a working Mutation object.
!

!MEMutationFactory class methodsFor: 'private' stamp: 'ag 7/21/2012 01:25:22.362'!
availableMutations
	
	^ MEMutation allSubclasses reject: #isAbstract! !

!MEMutationFactory class methodsFor: 'cleanup' stamp: 'ag 7/21/2012 01:24:58.772'!
cleanupMethodWrappers
	
	MEReplacementMethodWrapper
		allInstancesDo: #cleanup.! !

!MEMutationFactory class methodsFor: 'instance creation' stamp: 'ag 7/21/2012 01:24:40.355'!
new
	
	^ self withInjector: MEMutationInjector default! !

!MEMutationFactory class methodsFor: 'instance creation' stamp: 'ag 7/21/2012 01:24:40.374'!
onMethodWrappers
	
	^ self withInjector: MEMutationInjector onMethodWrappers! !

!MEMutationFactory class methodsFor: 'instance creation' stamp: 'ag 7/21/2012 01:24:40.365'!
onSourcecode
	
	^ self withInjector: MEMutationInjector onSourcecode! !

!MEMutationFactory class methodsFor: 'instance creation' stamp: 'ag 7/20/2012 14:13:07.779'!
withInjector: aMutationInjector
	
	^ self basicNew
		injector: aMutationInjector;
		availableMutations: self availableMutations;
		initialize;
		yourself! !

!MEMutationFactory methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:45:41.569'!
availableMutations
	^ availableMutations! !

!MEMutationFactory methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:45:41.6'!
availableMutations: anObject
	availableMutations := anObject! !

!MEMutationFactory methodsFor: 'mutation creation' stamp: 'ag 5/31/2012 15:14:42.011'!
createMutation: mutationClass
	"Random: the mutation-parameters, the mutated method.
	Explicit: the mutation-class."
	
	^ self
		createMutation: mutationClass
		forMethod:
			((self randomMethodForMutation: mutationClass)
				ifNil: [ ^ nil ])! !

!MEMutationFactory methodsFor: 'mutation creation' stamp: 'ag 7/20/2012 14:51:58.048'!
createMutation: mutationClass forMethod: method
	"Random: the mutation-parameters.
	Explicit: the mutation-class, the mutated method."
	
	| parameters |
	(mutationClass isAbleToMutate: (injector astOfMethod: method)) ifFalse: [ ^ nil ].
	parameters := mutationClass randomParametersFor: (injector astOfMethod: method).
	^ self
		createMutation: mutationClass
		withParameters: parameters
		forMethod: method! !

!MEMutationFactory methodsFor: 'mutation creation' stamp: 'ag 5/31/2012 15:18:28.794'!
createMutation: mutationClass withParameters: parameterDictionary
	"Random: the mutated method.
	Explicit: the mutation-class, the mutation-parameters."
	
	^ self
		createMutation: mutationClass
		withParameters: parameterDictionary
		forMethod:
			((self randomMethodForMutation: mutationClass withParameters: parameterDictionary)
				ifNil: [ ^ nil ])! !

!MEMutationFactory methodsFor: 'mutation creation' stamp: 'ag 7/20/2012 14:50:58.456'!
createMutation: mutationClass withParameters: parameterDictionary forMethod: method
	"Random: nothing.
	Explicit: the mutation-class, the mutation-parameters, the mutated method."
	
	(mutationClass isAbleToMutate: (injector astOfMethod: method) withParameters: parameterDictionary)
		ifFalse: [ ^ nil ].
	^ mutationClass
		injector: injector
		parameters: 
			(parameterDictionary
				at: #method put: method;
				yourself)! !

!MEMutationFactory methodsFor: 'private' stamp: 'ag 7/20/2012 14:24:05.69'!
detectRandom: aCollection satisfying: aBlock
	
	| collection |
	aCollection isEmpty ifTrue: [ ^ nil ].
	collection := aCollection deepCopy.
	[ collection isEmpty ] whileFalse: [ | element |
		element := collection atRandom.
		(aBlock value: element) == true "Allow non-boolean return values"
			ifTrue: [ ^ element ].
		collection remove: element ].
	^ nil! !

!MEMutationFactory methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:45:41.208'!
injector
	^ injector! !

!MEMutationFactory methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:45:41.487'!
injector: anObject
	injector := anObject! !

!MEMutationFactory methodsFor: 'private' stamp: 'ag 5/31/2012 15:47:26.323'!
randomMethodForMutation: aMutationClass
	"Return a random method from the MutationScope, that can be mutated by aMutationClass.
	Pick random methods, until one is found, that is applicable.
	Of none is applicable, return nil."
	
	^ self
		detectRandom: scope
		satisfying: [ :method |
			aMutationClass
				isAbleToMutate: (injector astOfMethod: method) ]! !

!MEMutationFactory methodsFor: 'private' stamp: 'ag 5/31/2012 15:47:22.517'!
randomMethodForMutation: aMutationClass withParameters: parameters
	"Return a random method from the MutationScope, that can be mutated by aMutationClass with the given parameters.
	Pick random methods, until one is found, that is applicable."
	
	^ self
		detectRandom: scope
		satisfying: [ :method |
				aMutationClass
					isAbleToMutate: (injector astOfMethod: method)
					withParameters: parameters ]! !

!MEMutationFactory methodsFor: 'mutation creation' stamp: 'ag 7/20/2012 14:49:35.223'!
randomMutation
	"Random: the mutation-class, the mutation-parameters, the mutated method.
	Explicit: nothing.
	
	Random mutation-class with explicit mutation-parameters is not supported, because it makes little
	sense to provide parameters for an unknown entity.
	
	This rather complex method iterates over all possible combinations of methods and mutations, trying to find a matching one."
	
	| mutation method |
	mutation := self
		detectRandom: availableMutations
		satisfying: [ :mutationClass |
			method := self
				detectRandom: scope
				satisfying: [ :checkedMethod | mutationClass isAbleToMutate: (injector astOfMethod: checkedMethod) ].
			method notNil ].
	mutation isNil | method isNil ifTrue: [ ^ nil ].
	^ self createMutation: mutation forMethod: method! !

!MEMutationFactory methodsFor: 'private' stamp: 'ag 5/31/2012 15:47:17.362'!
randomMutationClassForMethod: aMethod
	"Return a mutation-class from availableMutations, that is actually able to mutate aMethod.
	Pick random mutation-classes and check, whether they are applicable, until a match is found or all classes are checked.
	If no class matches the criteria, return nil."
	
	^ self
		detectRandom: availableMutations
		satisfying: [ :mutationClass |
			mutationClass
				isAbleToMutate: (injector astOfMethod: aMethod) ]! !

!MEMutationFactory methodsFor: 'mutation creation' stamp: 'ag 5/31/2012 15:25:37.272'!
randomMutationForMethod: method
	"Random: the mutation-class, the mutation-parameters.
	Explicit: the mutated method.
	
	Random mutation-class with explicit mutation-parameters is not supported, because it makes little
	sense to provide parameters for an unknown entity."
	
	^ self
		createMutation:
			((self randomMutationClassForMethod: method)
				ifNil: [ ^ nil ])
		forMethod: method! !

!MEMutationFactory methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:45:41.515'!
scope
	^ scope! !

!MEMutationFactory methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:45:41.543'!
scope: anObject
	scope := anObject! !

Object subclass: #MEMutationInjector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEMutationInjector commentStamp: 'ag 7/9/2012 12:23:43.29' prior: 0!
A MutationInjector provides services for Mutation objects.
The mutations are handled on AST-level. The Injector provides access to ASTs and methods to apply changed ASTs to mutated methods.

In addition, the current 'state' of the methods can be queried and restored later.
This is used to undo/revert the Mutation.
Depending on the Injector-implementation, different objects are used to represent the 'state' of the method. Some implementations do not need any state at all, which might be preferrable in terms of simplicity/safety. These should subclass from StatelessMutationInjector.

Different Injector-implementations (subclasses) apply the modified AST on different levels.
E.g. a SourcecodeMutationInjector recompiles the modified source-code (obtained from the modified AST) inside the class.

Other ideas for Injector-implementations:
 - BytecodeMutationInjector - compiled the modified source-code and (somehow) set the bytecode directly into the CompiledMethod. Dont know wheterh thats possible, but the idea is that cleaning up is easier, because the source-code of the method has not changed, and we simply have to recompile it.
 - ContextMutationInjector - use ContextS or ContextS2. Create a Layer for Mutations, add a layered version of the method, with the modified source-code, after the mutation, simply delete the layer
!

!MEMutationInjector class methodsFor: 'instance creation' stamp: 'ag 7/9/2012 11:38:21.099'!
default
	"MutationInjector objects are stateless, but create new instances here nonetheless. Avoids class-variables.
	Also, state might be introduced in the future."
	
	^ self onSourcecode! !

!MEMutationInjector class methodsFor: 'instance creation' stamp: 'ag 7/21/2012 01:25:16.809'!
onMethodWrappers
	
	^ MEMethodWrapperMutationInjector new! !

!MEMutationInjector class methodsFor: 'instance creation' stamp: 'ag 7/21/2012 01:25:06.047'!
onSourcecode
	
	^ MESourcecodeMutationInjector new! !

!MEMutationInjector methodsFor: 'AST manipulation' stamp: 'ag 7/9/2012 11:19:32.174'!
applyAst: parseTree toMethod: method
	"parseTree: RBMethodNode, method: class -> selector-symbol.
	Conrete subclasses decide, how the modified AST of a method is applied into the system.
	The object returned here will be stored by the invoking Mutation object. When reverting the Mutation,
	revertingAfterHavingApplied: will be invoked with this object as argument."
	
	self subclassResponsibility.! !

!MEMutationInjector methodsFor: 'AST manipulation' stamp: 'ag 5/25/2012 18:05:54.559'!
astOfMethod: method
	"method: class -> selector-symbol"
	
	^ (method key >> method value) parseTree! !

!MEMutationInjector methodsFor: 'state restoring' stamp: 'ag 5/25/2012 13:43:03.789'!
currentStateOfMethod: method
	"method: class -> selector-symbol.
	Subclasses can return an arbitrary object here, that has later ot be used by #restoreStateOfMethod:to:
	to restore the method to its original state."

	self subclassResponsibility.! !

!MEMutationInjector methodsFor: 'state restoring' stamp: 'ag 5/25/2012 13:43:09.929'!
restoreStateOfMethod: method to: anObject
	"method: class -> selector-symbol.
	anObject: return value of the according #currentStateOfMethod: call. See the comment there."

	self subclassResponsibility.! !

!MEMutationInjector methodsFor: 'state restoring' stamp: 'ag 7/9/2012 11:18:45.772'!
revertingAfterHavingApplied: injectorState
	"Do nothing, subclasses may include additional actions.
	This will be invoked, when a Mutation is reverted. injectorState is the object, that was previously returned by
	applyAst:toMethod:"
	! !

MEMutationInjector subclass: #MESourcecodeMutationInjector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MESourcecodeMutationInjector commentStamp: 'ag 7/9/2012 12:27:59.26' prior: 0!
This MutationInjector compiles the mutated sourcecode directly into the mutated class.
This modifies the global behaviour of the mutated method in the image.
If something prevents the Mutation from cleanly reverting itself (e.g., the image crashes), this might permanently corrupt the image
and require the user to fix the state.
If using this injector-implementation, the MutationScope should be chosen carefully to not include any System-internal methods and classes.
This MutationInjector represents the state of a method by the complete method-sourcecode.!

!MESourcecodeMutationInjector methodsFor: 'AST manipulation' stamp: 'ag 5/25/2012 18:28:35.864'!
applyAst: parseTree toMethod: method
	"Get the source from the parseTree, and recompile the source inside the class.
	Changes made to the tree try to preseve the structure of the original code."
	
	self applyCode: parseTree newSource toMethod: method.! !

!MESourcecodeMutationInjector methodsFor: 'AST manipulation' stamp: 'ag 5/25/2012 14:13:17.967'!
applyCode: sourceCode toMethod: method
	"Find the category of the method, and recompile the given code inside the class.
	Trust in the fact, that the code actually compiles to the selector given in the method-parameter."
	
	method key
		compile: sourceCode
		classified:
			(method key organization
				categoryOfElement: method value).! !

!MESourcecodeMutationInjector methodsFor: 'state restoring' stamp: 'ag 5/25/2012 18:08:35.792'!
currentStateOfMethod: method
	"The 'state' of this MutationInjector-implementation is the source-code of the method.'"
	
	^ (method key >> method value) getSource! !

!MESourcecodeMutationInjector methodsFor: 'state restoring' stamp: 'ag 5/25/2012 14:14:01.494'!
restoreStateOfMethod: method to: anObject
	"anObject is the source-code of method, as returned from currentStateOfMethod:.
	Recompile this original code."
	
	self applyCode: anObject toMethod: method.! !

MEMutationInjector subclass: #MEStatelessMutationInjector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEStatelessMutationInjector commentStamp: 'ag 7/9/2012 12:27:28.407' prior: 0!
This is an abstract base-class for MutationInjector implementations, that do not provide meaningfull
implementations for currentStateOfMethod: and restoreStateOfMethod:to:.
!

MEStatelessMutationInjector subclass: #MEMethodWrapperMutationInjector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEMethodWrapperMutationInjector commentStamp: 'ag 7/9/2012 12:33:46.29' prior: 0!
This MutationInjector wraps the mutated method with a special MwMethodWrapper (ReplacementMethodWrapper).
The mutated source-code is compiled anonymously, without being inserted into the method-dictionary.
The resulting "anonymous" instance of CompiledMethod is executed by the MethodWrapper instead of the original method.

This implemention is safer than the SourcecodeMutationInjector, because the original version of the mutated method is still present in the
image, and is not completely replaced. When a Mutation seems to not have cleanly reverted itself (e.g. because of an image-crash),
the method "MethodWrapperMutationInjector nuke" can be used to delete/uninstall any leftover instances of ReplacementMethodWrapper.

SourcecodeMutationInjector should have a big advantage regarding performance, which is most likely not relevant when using the MutationEngine library.!

!MEMethodWrapperMutationInjector methodsFor: 'AST manipulation' stamp: 'ag 7/21/2012 01:24:58.781'!
applyAst: parseTree toMethod: method
	"Code to compile the method taken (and modified) from Behavior>>#compile:notifying:"
	
	| methodAndNode wrapper |
	methodAndNode  := method key
		compile: parseTree newSource
		classified: nil
		notifying: nil
		trailer: method key defaultMethodTrailer
		ifFail: [ self error:
			'Could not compile modified code of mutated method!! Inspect the code over ''parseTree newSource''.' ].
	wrapper := MEReplacementMethodWrapper
		replace: method value
		inClass: method key
		with: methodAndNode method.
	wrapper install.
	^ wrapper! !

!MEMethodWrapperMutationInjector methodsFor: 'state restoring' stamp: 'ag 7/9/2012 11:22:18.921'!
revertingAfterHavingApplied: wrapper
	
	wrapper uninstall.! !

!MEStatelessMutationInjector methodsFor: 'state restoring' stamp: 'ag 5/25/2012 14:37:45.545'!
currentStateOfMethod: method
	"The object returned here does not actually matter, just don't return nil, since Mutation uses nil to mark,
	that the state has not yet been stored.
	#restoreStateOfMethod:to: still has to be implemented of subclasses, but the parameter-object is 'meaningless'."
	
	^ #NoState! !

!MEStatelessMutationInjector methodsFor: 'state restoring' stamp: 'ag 7/9/2012 11:21:57.955'!
restoreStateOfMethod: method to: anObject
	"Do nothing, since no meaningful value is returned by currentStateOfMethod:"
	! !

Object subclass: #MEMutationScope
	instanceVariableNames: 'methods'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEMutationScope commentStamp: 'ag 6/17/2012 18:44:53.395' prior: 0!
A MutationScrope is a collection of Methods (represented as a Dictionary from Class to Set<Symbol>).
There are many convenience methods, that allow quick and convenient modification of this collection,
to quickly describe the scope, that should be used by a MutationFactory, when selecting random methods to mutate.
An instance of MutationScope should be set into every instance of MutationFactory (when the random selection of methods is required).
A MutationScope implements some methods from the Collection-interface.!

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 8/8/2012 14:06:38.136'!
add: anObject
	"Convenience-method, allowing to add different kinds of objects to the MutationScope.
	(This is also part of the Collection interface)."
	
	anObject class == MethodReference ifTrue: [ ^ self addMethodReference: anObject ].
	anObject isBehavior ifTrue: [ ^ self addClass: anObject ].
	anObject isCompiledMethod ifTrue: [ ^ self addCompiledMethod: anObject ].
	anObject class == Association ifTrue: [ ^ self addMethod: anObject ].
	anObject isDictionary ifTrue: [ ^ self addMethodsFromDictionary: anObject ].
	(anObject isCollection and: anObject notEmpty) ifTrue: [ | probe | "Should contain homogenous elements..."
		probe := anObject anyOne.
		probe isBehavior ifTrue: [ ^ self addClasses: anObject ].
		probe isCompiledMethod ifTrue: [ anObject do: [ :method | self addCompiledMethod: method ]. ^ self ].
		probe class == Association ifTrue: [ | dict |
			dict := Dictionary new.
			anObject do: [ :assoc | (dict at: assoc key ifAbsentPut: Set new) add: assoc value ].
			^ self addMethodsFromDictionary: dict ] ].
	self error: 'Don''t know how to add this object to this mutation-scope.'.! !

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 7/20/2012 15:19:40.34'!
addClass: aClass
	
	self addMethods: aClass organization allMethodSelectors forClass: aClass.! !

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:12:03.417'!
addClasses: setOfClasses
	
	setOfClasses do: [ :cl | self addClass: cl ].! !

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 14:57:45.378'!
addCompiledMethod: aCompiledMethod
	
	self addMethod: (aCompiledMethod methodClass -> aCompiledMethod selector).! !

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 14:59:42.146'!
addMethod: association
	"Class -> selector"
	
	self addMethods: { association value } forClass: association key.! !

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 8/8/2012 14:06:00.006'!
addMethodReference: aMethodReference
	
	self addMethods: { aMethodReference selector } forClass: aMethodReference actualClass! !

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:14:21.383'!
addMethods: selectors forClass: aClass
	
	| methodSet |
	methodSet := methods at: aClass ifAbsentPut: [ Set new ].
	methodSet addAll: selectors.! !

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:13:19.294'!
addMethodsFromDictionary: aDictionary
	"Class -> (Set of selector-object). Adds only the specified methods for each class."
	
	aDictionary keysAndValuesDo: [ :class :methodSet |
		self addMethods: methodSet forClass: class ].! !

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 7/21/2012 00:36:16.247'!
addPackage: aPackageName
	"Add all classes and all methods in the given package."
	
	self addClasses:
		((SystemOrganizer default listAtCategoryNamed: aPackageName)
			collect: [ :className | Smalltalk at: className asSymbol ]).! !

!MEMutationScope methodsFor: 'method selection' stamp: 'ag 7/21/2012 00:59:16.475'!
allMethods
	
	^ methods keys
		inject: OrderedCollection new
		into: [ :all :key | 
			(methods at: key)
				do: [ :value | all add: key -> value ].
			all ]! !

!MEMutationScope methodsFor: 'method selection' stamp: 'ag 5/31/2012 15:02:47.451'!
atRandom
	"Be compliant to the Collection interface."
	
	^ self randomMethod! !

!MEMutationScope methodsFor: 'organization' stamp: 'ag 7/20/2012 14:18:04.985'!
cleanup
	"Remove all classes from the dictionary, that have no methods associated to them.
	This should rarely be necessary, but is provided for completeness."
	
	| deadClasses |
	deadClasses := Set new.
	methods keysAndValuesDo: [ :class :selectorSet |
		selectorSet ifEmpty: [ deadClasses add: class ] ].
	deadClasses do: [ :deadClass | methods removeKey: deadClass ].! !

!MEMutationScope methodsFor: 'initialization' stamp: 'ag 5/25/2012 15:09:17.878'!
initialize
	
	super initialize.
	methods := Dictionary new. "Dictionary from Class to (Set of Symbol)"! !

!MEMutationScope methodsFor: 'testing' stamp: 'ag 7/20/2012 14:24:58.488'!
isEmpty
	"Be compliant to the Collection interface."
	
	self cleanup.
	^ methods isEmpty! !

!MEMutationScope methodsFor: 'method selection' stamp: 'ag 7/20/2012 14:25:08.613'!
randomMethod
	"Returns association - Class -> Selector-Symbol. Returns nil, if this scope is empty."
	
	| class |
	self isEmpty ifTrue: [ ^ nil ].
	class := methods keys atRandom.
	^ class -> (methods at: class) atRandom! !

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/31/2012 15:04:57.797'!
remove: anObject
	"(This is also part of the Collection interface)"
	
	anObject isBehavior ifTrue: [ ^ self removeClass: anObject ].
	anObject isCompiledMethod ifTrue: [ ^ self removeCompiledMethod: anObject ].
	(anObject isKindOf: Association) ifTrue: [ ^ self removeMethod: anObject ].
	anObject isDictionary ifTrue: [ ^ self removeAll: anObject ].
	self error: 'Don''t know how to handle object, passed for removal!!'.! !

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:17:51.449'!
removeAll: aDictionary
	"Class -> (Set of selector-symbols). Remove all methods found in that dictionary."
	
	aDictionary keysAndValuesDo: [ :class :selectorSet |
		methods
			at: class
			ifPresent: [ :currentSelectorSet | currentSelectorSet removeAll: selectorSet ]
			ifAbsent: [ "Nothing, continue" ] ].! !

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:15:19.809'!
removeClass: aClass
	"Removes all methods in that class."
	
	methods remove: aClass ifAbsent: [ ^ self ].! !

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:25:48.222'!
removeCompiledMethod: aCompiledMethod
	
	self removeMethod: aCompiledMethod methodClass -> aCompiledMethod selector.! !

!MEMutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:09:01.58'!
removeMethod: association
	"Class -> selector-symbol. The class is not removed, even if all methods for that class are removed."
	
	methods
		at: association key
		ifPresent: [ :set | set remove: association value ifAbsent: [ ^ self ] ]
		ifAbsent: [ ^ self ].! !

!MEMutationScope methodsFor: 'accessing' stamp: 'ag 7/21/2012 00:39:00.308'!
size
	"Count all methods contained in this mutation-scope. Implement Collection interface."
	
	^ methods inject: 0 into: [ :all :valueSet | all + valueSet size ]! !

!MEMutationScope methodsFor: 'method selection' stamp: 'ag 7/20/2012 15:22:03.964'!
uniqueRandomMethod
	"Return a random method and remove that method from this scope.
	This way, the scope can be fully traversed, wihtout hitting duplicates.
	The receiver can be deepCopied before doing that, if it needs to be reused."
	
	| method |
	self isEmpty ifTrue: [ ^ nil ].
	method := self randomMethod.
	self removeMethod: method.
	^ method! !

Object subclass: #MEMutationTestObjectHelper
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine-Tests'!

!MEMutationTestObjectHelper class methodsFor: 'tests' stamp: 'ag 7/21/2012 00:33:10.792'!
dummyMethod! !

!MEMutationTestObjectHelper class methodsFor: 'tests' stamp: 'ag 7/9/2012 11:59:46.127'!
otherValue
	
	^ 'helper-value'! !

MwMethodWrapper subclass: #MEReplacementMethodWrapper
	instanceVariableNames: 'replacementMethod'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MEReplacementMethodWrapper commentStamp: 'ag 7/9/2012 12:35:54.172' prior: 0!
This specialization of MwMethodWrapper is used to implement MethodWrapperMutationInjector.
It completely replaces the wrapped method, instead of executing additional code before/after the original method.
Also, an instance of CompiledMethod is used to represent the replacement-behavior (instead of a Block).!

!MEReplacementMethodWrapper class methodsFor: 'as yet unclassified' stamp: 'ag 7/9/2012 11:29:55.856'!
replace: aSelector inClass: aClass with: otherCompiledMethod
	
	^ (self
		on: aSelector
		inClass: aClass)
		replacementMethod: otherCompiledMethod;
		yourself! !

!MEReplacementMethodWrapper methodsFor: 'cleanup' stamp: 'ag 7/17/2012 15:04:15.266'!
cleanup
	
	self isInstalled ifTrue: [
		[ self uninstall ] on: Error do: [ "Silent" ] ]! !

!MEReplacementMethodWrapper methodsFor: 'accessing' stamp: 'ag 7/9/2012 11:30:03.937'!
replacementMethod
	^ replacementMethod! !

!MEReplacementMethodWrapper methodsFor: 'accessing' stamp: 'ag 7/9/2012 11:30:03.971'!
replacementMethod: anObject
	replacementMethod := anObject! !

!MEReplacementMethodWrapper methodsFor: 'evaluating' stamp: 'ag 7/9/2012 11:30:25.866'!
valueWithReceiver: receiver arguments: args
	
	^ replacementMethod valueWithReceiver: receiver arguments: args! !

TestCase subclass: #MEMutationOtherTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine-Tests'!

!MEMutationOtherTests methodsFor: 'tests' stamp: 'ag 7/21/2012 01:28:39.538'!
testAvailableMutations
	
	| mutations |
	mutations := MEMutationFactory availableMutations.
	self
		deny: mutations includes: MEMutation;
		deny: mutations includes: MEAstMutation;
		deny: mutations includes: METokenMutation;
		deny: mutations includes: MEKeywordArgumentMutation;
		assert: mutations includes: MEStringLiteralMutation;
		assert: mutations includes: MEAddAssignmentMessageMutation;
		assert: mutations includes: MEAddCopyAssignmentMutation;
		assert: mutations includes: MESelfSuperMutation;
		assert: mutations includes: MEIdentifierMutation.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.145'!
testMutationScope1
	
	self assert: MEMutationScope new isEmpty! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.073'!
testMutationScope10
	
	| scope methods |
	methods := {
		MEMutationTestObject class -> #returnNil.
		MEMutationTestObject class -> #returnBoolean.
		MEMutationTestObjectHelper class -> #otherValue.
		MEMutationTestObjectHelper class -> #dummyMethod.
	}.
	scope := MEMutationScope new.
	methods do: [ :assoc | scope addMethod: assoc ].
	self assert: methods asSet equals: scope allMethods asSet.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.042'!
testMutationScope11
	
	| scope methods |
	scope := MEMutationScope new.
	scope addClasses: { MEMutationTestObject class. MEMutationTestObjectHelper class }.
	methods := MEMutationTestObject class organization allMethodSelectors collect: [ :selector | MEMutationTestObject class -> selector ].
	methods := methods asOrderedCollection.
	methods addAll: (MEMutationTestObjectHelper class organization allMethodSelectors collect: [ :selector | MEMutationTestObjectHelper class -> selector ]).
	self assert: methods asSet equals: scope allMethods asSet.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.032'!
testMutationScope12
	
	| scope methods |
	scope := MEMutationScope new.
	scope addClass: MEMutationTestObject class.
	methods := MEMutationTestObject class organization allMethodSelectors collect: [ :selector | MEMutationTestObject class -> selector ].
	self assert: methods asSet equals: scope allMethods asSet.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.01'!
testMutationScope13
	
	| scope methods |
	scope := MEMutationScope new.
	scope add: MEMutationTestObject class.
	methods := MEMutationTestObject class organization allMethodSelectors collect: [ :selector | MEMutationTestObject class -> selector ].
	self assert: methods asSet equals: scope allMethods asSet.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.175'!
testMutationScope14
	
	| scope methods |
	methods := {
		MEMutationTestObject class -> #returnNil.
		MEMutationTestObject class -> #returnBoolean.
		MEMutationTestObjectHelper class -> #otherValue.
		MEMutationTestObjectHelper class -> #dummyMethod.
	}.
	scope := MEMutationScope new.
	methods do: [ :assoc | scope addCompiledMethod: assoc key >> assoc value ].
	self assert: methods asSet equals: scope allMethods asSet.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.154'!
testMutationScope15
	
	| scope methods |
	methods := {
		MEMutationTestObject class -> #returnNil.
		MEMutationTestObject class -> #returnBoolean.
		MEMutationTestObjectHelper class -> #otherValue.
		MEMutationTestObjectHelper class -> #dummyMethod.
	}.
	scope := MEMutationScope new.
	methods do: [ :assoc | scope add: assoc key >> assoc value ].
	self assert: methods asSet equals: scope allMethods asSet.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.134'!
testMutationScope16
	
	| scope methods |
	methods := {
		MEMutationTestObject class -> #returnNil.
		MEMutationTestObject class -> #returnBoolean.
		MEMutationTestObjectHelper class -> #otherValue.
		MEMutationTestObjectHelper class -> #dummyMethod.
	}.
	scope := MEMutationScope new.
	methods do: [ :assoc | scope add: assoc ].
	self assert: methods asSet equals: scope allMethods asSet.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.115'!
testMutationScope17
	
	| scope methods |
	methods := {
		MEMutationTestObject class -> #returnNil.
		MEMutationTestObject class -> #returnBoolean.
		MEMutationTestObjectHelper class -> #otherValue.
		MEMutationTestObjectHelper class -> #dummyMethod.
	}.
	scope := MEMutationScope new.
	scope add: methods.
	self assert: methods asSet equals: scope allMethods asSet.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.094'!
testMutationScope18
	
	| scope methods |
	methods := {
		MEMutationTestObject class -> #returnNil.
		MEMutationTestObject class -> #returnBoolean.
		MEMutationTestObjectHelper class -> #otherValue.
		MEMutationTestObjectHelper class -> #dummyMethod.
	}.
	scope := MEMutationScope new.
	scope add: (methods collect: [ :assoc | assoc key >> assoc value ]).
	self assert: methods asSet equals: scope allMethods asSet.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.063'!
testMutationScope19
	
	| scope |
	scope := MEMutationScope new.
	scope addPackage: 'MutationEngine-Tests'.
	(SystemOrganizer default listAtCategoryNamed: 'MutationEngine-Tests')
		do: [ :className | | cl |
			cl := Smalltalk at: className.
			cl organization allMethodSelectors
				do: [ :selector |
					scope remove: cl -> selector ] ].
	self assert: scope isEmpty.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.125'!
testMutationScope2
	
	self assert: MEMutationScope new atRandom isNil! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.104'!
testMutationScope3
	
	| scope expectedMethods methods |
	expectedMethods := { MEMutationTestObject class -> #returnStringOne. MEMutationTestObject class -> #returnStringThree }.
	scope := MEMutationScope new.
	expectedMethods do: [ :mt | scope add: mt ].
	methods := Set with: scope uniqueRandomMethod.
	self assert: expectedMethods size - 1 equals: scope size.
	methods addAll: scope allMethods.
	self assert: expectedMethods asSet equals: methods.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.084'!
testMutationScope4
	
	| scope |
	scope := MEMutationScope new.
	scope addMethod: MEMutationTestObject class -> #returnNil.
	self deny: scope isEmpty.
	scope removeMethod: MEMutationTestObject class -> #returnNil.
	self assert: scope isEmpty.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.053'!
testMutationScope5
	
	| scope |
	scope := MEMutationScope new.
	scope addMethod: MEMutationTestObject class -> #returnNil.
	self deny: scope isEmpty.
	scope cleanup. "Should not delete anything."
	self deny: scope isEmpty.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.02'!
testMutationScope7
	
	| scope allMethods |
	scope := MEMutationScope new.
	scope addPackage: 'MutationEngine-Tests'.
	allMethods := scope allMethods.
	(SystemOrganizer default listAtCategoryNamed: 'MutationEngine-Tests')
		do: [ :className | | cl |
			cl := Smalltalk at: className.
			cl organization allMethodSelectors
				do: [ :selector |
					allMethods remove: cl -> selector ] ].
	self assert: allMethods isEmpty.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.185'!
testMutationScope8
	
	| scope methods methodsDict |
	methods := {
		MEMutationTestObject class -> #returnNil.
		MEMutationTestObject class -> #returnBoolean.
		MEMutationTestObjectHelper class -> #otherValue.
		MEMutationTestObjectHelper class -> #dummyMethod.
	}.
	scope := MEMutationScope new.
	methodsDict := Dictionary new.
	methods do: [ :assoc | (methodsDict at: assoc key ifAbsentPut: Set new) add: assoc value ].
	scope addMethodsFromDictionary: methodsDict.
	self assert: methods asSet equals: scope allMethods asSet.! !

!MEMutationOtherTests methodsFor: 'tests-MutationScope' stamp: 'ag 7/21/2012 01:24:53.164'!
testMutationScope9
	
	| scope methods methodsDict |
	methods := {
		MEMutationTestObject class -> #returnNil.
		MEMutationTestObject class -> #returnBoolean.
		MEMutationTestObjectHelper class -> #otherValue.
		MEMutationTestObjectHelper class -> #dummyMethod.
	}.
	methodsDict := Dictionary new.
	methods do: [ :assoc | (methodsDict at: assoc key ifAbsentPut: Set new) add: assoc value ].
	scope := MEMutationScope new.
	methodsDict keysAndValuesDo: [ :key :value |
		scope addMethods: value forClass: key ].
	self assert: methods asSet equals: scope allMethods asSet.! !

TestCase subclass: #MEMutationTestObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine-Tests'!

!MEMutationTestObject class methodsFor: 'tests' stamp: 'ag 7/5/2012 15:30:54.505'!
otherValue
	
	^ 'super-value'! !

!MEMutationTestObject class methodsFor: 'tests' stamp: 'mp 7/23/2012 14:59:47.754'!
returnBoolean
	
	^ true! !

!MEMutationTestObject class methodsFor: 'tests' stamp: 'mp 7/23/2012 14:59:47.709'!
returnNil
	
	^ nil! !

!MEMutationTestObject class methodsFor: 'tests' stamp: 'mp 7/23/2012 14:59:48.11'!
returnNumberOne
	
	^ 222! !

!MEMutationTestObject class methodsFor: 'tests' stamp: 'mp 7/23/2012 14:59:48.062'!
returnNumberThree
	
	| result other |
	other := 111.
	result := 222.
	other := 333.
	^ result! !

!MEMutationTestObject class methodsFor: 'tests' stamp: 'mp 7/23/2012 14:59:48.353'!
returnStringOne
	
	^ 'a-string'! !

!MEMutationTestObject class methodsFor: 'tests' stamp: 'mp 7/23/2012 14:59:48.305'!
returnStringThree
	
	| result other |
	other := 'wrong-string-1'.
	result := 'a-string'.
	other := 'wrong-string-2'.
	^ result! !

MEMutationTestObject subclass: #MEMutationTestSubObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine-Tests'!

MEMutationTestSubObject class
	instanceVariableNames: 'testObj'!

MEMutationTestSubObject class
	instanceVariableNames: 'testObj'!

!MEMutationTestSubObject class methodsFor: 'initialize - event' stamp: 'ag 7/9/2012 12:00:54.068'!
initialize
	"self initialize"
	
	super initialize.
	testObj := self new.! !

!MEMutationTestSubObject class methodsFor: 'tests' stamp: 'ag 7/5/2012 15:37:37.549'!
otherValue
	
	^ 'sub-value'! !

!MEMutationTestSubObject class methodsFor: 'tests' stamp: 'mp 7/23/2012 14:59:47.437'!
returnName
	
	^ self name! !

!MEMutationTestSubObject class methodsFor: 'tests' stamp: 'mp 7/23/2012 14:59:47.483'!
returnOtherValue
	
	^ self otherValue! !

!MEMutationTestSubObject class methodsFor: 'tests' stamp: 'mp 7/23/2012 14:59:47.288'!
returnSuperOtherValue
	
	^ super otherValue! !

!MEMutationTestSubObject methodsFor: 'tests' stamp: 'ag 7/9/2012 11:53:31.468'!
otherValue
	
	^ 'sub-inst-value'! !

TestCase subclass: #MEMutationTests
	instanceVariableNames: 'mutation parameters'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine-Tests'!

!MEMutationTests methodsFor: 'running' stamp: 'ag 7/21/2012 01:24:40.383'!
createInjector
	"We conduct the normal tests just with the default injector.
	The other injectors are tested separately.
	Since the Mutations just use the API of the injector, this should be enough."
	
	^ MEMutationInjector default! !

!MEMutationTests methodsFor: 'running' stamp: 'ag 7/9/2012 11:33:50.781'!
injector: injector mutation: mutationClass method: methodAssociation nonMutated: normalBlock parameters: params do: testBlock
	
	| parameters |
	"First test with generated prameters."
	parameters := params as: Dictionary.
	parameters at: #method put: methodAssociation.
	mutation := mutationClass injector: injector parameters: parameters.
	self deny: mutation isApplied.
	normalBlock value.
	mutation doMutated: [
		self assert: mutation isApplied.
		testBlock value ].
	self deny: mutation isApplied.
	normalBlock value.! !

!MEMutationTests methodsFor: 'running' stamp: 'ag 7/9/2012 11:37:32.1'!
mutation: mutationClass method: methodAssociation nonMutated: normalBlock parameters: params do: testBlock
	
	^ self
		injector: self createInjector
		mutation: mutationClass
		method: methodAssociation
		nonMutated: normalBlock
		parameters: params
		do: testBlock! !

!MEMutationTests methodsFor: 'running' stamp: 'ag 7/5/2012 13:56:00.167'!
randomMutation: mutationClass method: methodAssociation nonMutated: normalBlock do: testBlock
	"This tests the mutation with randomly created parameters.
	This is not very nicely reproduceable. When such a test fails, it is very important to inspect the create mutation-object
	(especially the contained parameters) and check, which parameters caused the unexpected behaviour.
	Also, such a test will most likely be only able to test, that something HAS changed, not WHAT exactly."
	
	| injector parameters |
	injector := self createInjector.
	parameters := mutationClass randomParametersFor: (injector astOfMethod: methodAssociation).
	parameters at: #method put: methodAssociation.
	mutation := mutationClass injector: injector parameters: parameters.
	self deny: mutation isApplied.
	normalBlock value.
	mutation doMutated: [
		self assert: mutation isApplied.
		testBlock value ].
	self deny: mutation isApplied.
	normalBlock value.! !

!MEMutationTests methodsFor: 'running' stamp: 'ag 7/9/2012 12:06:21.905'!
tearDown
	
	| wasApplied |
	wasApplied := mutation isApplied.
	wasApplied ifTrue: [ mutation revert ].
	self deny: wasApplied.! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:26:40.957'!
testIdentifierMutation1
	
	self
		mutation: MEIdentifierMutation
		method: MEMutationTestSubObject class -> #returnSuperOtherValue
		nonMutated: [ self assert: 'super-value' equals: MEMutationTestSubObject returnSuperOtherValue ]
		parameters: {
			#replacementMode -> #useSelf.
		}
		do: [ self assert: 'sub-value' equals: MEMutationTestSubObject returnSuperOtherValue ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:26:40.978'!
testIdentifierMutation2
	
	self
		mutation: MEIdentifierMutation
		method: MEMutationTestSubObject class -> #returnOtherValue
		nonMutated: [ self assert: 'sub-value' equals: MEMutationTestSubObject returnOtherValue ]
		parameters: {
			#replacementMode -> #useSuper.
		}
		do: [ self assert: 'super-value' equals: MEMutationTestSubObject returnOtherValue ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:26:40.947'!
testIdentifierMutation3
	
	self
		mutation: MEIdentifierMutation
		method: MEMutationTestSubObject class -> #returnOtherValue
		nonMutated: [ self assert: 'sub-value' equals: MEMutationTestSubObject returnOtherValue ]
		parameters: {
			#replacementMode -> #randomInstanceVariable.
		}
		do: [ self assert: 'sub-inst-value' equals: MEMutationTestSubObject returnOtherValue ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:26:40.968'!
testIdentifierMutation4
	
	self
		mutation: MEIdentifierMutation
		method: MEMutationTestSubObject class -> #returnName
		nonMutated: [ self assert: #MEMutationTestSubObject equals: MEMutationTestSubObject returnName ]
		parameters: {
			#replacementMode -> #randomClass.
		}
		do: [
			self deny: MEMutationTestSubObject returnName = #MEMutationTestSubObject.
			self assert: (Smalltalk classNamed: MEMutationTestSubObject returnName) notNil ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'mp 7/23/2012 14:59:42.771'!
testIdentifierMutation5
	
	self
		mutation: MEIdentifierMutation
		method: MEMutationTestSubObject class -> #returnOtherValue
		nonMutated: [ self assert: 'sub-value' equals: MEMutationTestSubObject returnOtherValue ]
		parameters: {
			#replacementMode -> #replace.
			#modificationValue -> 'MEMutationTestObjectHelper'.
		}
		do: [ self assert: 'helper-value' equals: MEMutationTestSubObject returnOtherValue ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:27.552'!
testKeywordMutation1
	
	self
		mutation: MEKeywordMutation
		method: MEMutationTestObject class -> #returnBoolean
		nonMutated: [ self assert: true equals: MEMutationTestObject returnBoolean ]
		parameters: {
			#replacementMode -> #preserveType.
		}
		do: [ self assert: false equals: MEMutationTestObject returnBoolean ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:27.592'!
testKeywordMutation2
	
	self
		mutation: MEKeywordMutation
		method: MEMutationTestObject class -> #returnNil
		nonMutated: [ self assert: nil equals: MEMutationTestObject returnNil ]
		parameters: {
			#replacementMode -> #preserveType.
		}
		do: [ self assert: nil equals: MEMutationTestObject returnNil ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:27.542'!
testKeywordMutation3
	
	self
		mutation: MEKeywordMutation
		method: MEMutationTestObject class -> #returnBoolean
		nonMutated: [ self assert: true equals: MEMutationTestObject returnBoolean ]
		parameters: {
			#replacementMode -> #random.
		}
		do: [ self assert: #(nil false) includes: MEMutationTestObject returnBoolean ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:27.582'!
testKeywordMutation4
	
	self
		mutation: MEKeywordMutation
		method: MEMutationTestObject class -> #returnNil
		nonMutated: [ self assert: nil equals: MEMutationTestObject returnNil ]
		parameters: {
			#replacementMode -> #random.
		}
		do: [ self assert: #(true false) includes: MEMutationTestObject returnNil ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:27.562'!
testKeywordMutationRandom
	
	self
		randomMutation: MEKeywordMutation
		method: MEMutationTestObject class -> #returnNil
		nonMutated: [ self assert: nil equals: MEMutationTestObject returnNil ]
		do: [ self assert: #(true false nil) includes: MEMutationTestObject returnNil ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:27.572'!
testKeywordMutationRandom2
	
	self
		randomMutation: MEKeywordMutation
		method: MEMutationTestObject class -> #returnBoolean
		nonMutated: [ self assert: true equals: MEMutationTestObject returnBoolean ]
		do: [ self assert: #(nil false) includes: MEMutationTestObject returnBoolean ]! !

!MEMutationTests methodsFor: 'tests - MethodWrapperInjector' stamp: 'ag 7/21/2012 01:28:39.591'!
testMwInjector
	"Test the Mw-injector by using one sample Mutation. It is also tested elsewhere."
	
	self assert: (MEMutationTestObject class >> #returnStringThree) class == CompiledMethod.
	self
		injector: MEMutationInjector onMethodWrappers
		mutation: MEStringLiteralMutation
		method: MEMutationTestObject class -> #returnStringThree
		nonMutated: [ self assert: 'a-string' equals: MEMutationTestObject returnStringThree ]
		parameters: {
			#nodeSelection -> 2.
			#replacementMode -> 'mutatedString'
		}
		do: [ self assert: 'mutatedString' equals: MEMutationTestObject returnStringThree ].
	self assert: (MEMutationTestObject class >> #returnStringThree) class == CompiledMethod.! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:33.33'!
testNumberLiteralMutation1
	
	self
		mutation: MENumberLiteralMutation
		method: MEMutationTestObject class -> #returnNumberThree
		nonMutated: [ self assert: 222 equals: MEMutationTestObject returnNumberThree ]
		parameters: {
			#nodeSelection -> 2.
			#replacementMode -> 4823.
			#nodeSelectionByIndex -> true.
		}
		do: [ self assert: 4823 equals: MEMutationTestObject returnNumberThree ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:33.37'!
testNumberLiteralMutation2
	
	self
		mutation: MENumberLiteralMutation
		method: MEMutationTestObject class -> #returnNumberThree
		nonMutated: [ self assert: 222 equals: MEMutationTestObject returnNumberThree ]
		parameters: {
			#nodeSelection -> 2.
			#replacementMode -> #random.
			#nodeSelectionByIndex -> true.
		}
		do: [ self assert: 222 ~= MEMutationTestObject returnNumberThree ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:33.319'!
testNumberLiteralMutation3
	
	self
		mutation: MENumberLiteralMutation
		method: MEMutationTestObject class -> #returnNumberThree
		nonMutated: [ self assert: 222 equals: MEMutationTestObject returnNumberThree ]
		parameters: {
			#nodeSelection -> 2.
			#replacementMode -> #smaller.
			#nodeSelectionByIndex -> true.
		}
		do: [ self assert: 222 > MEMutationTestObject returnNumberThree ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:33.36'!
testNumberLiteralMutation4
	
	self
		mutation: MENumberLiteralMutation
		method: MEMutationTestObject class -> #returnNumberThree
		nonMutated: [ self assert: 222 equals: MEMutationTestObject returnNumberThree ]
		parameters: {
			#nodeSelection -> 2.
			#replacementMode -> #bigger.
			#nodeSelectionByIndex -> true.
		}
		do: [ self assert: 222 < MEMutationTestObject returnNumberThree ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:33.309'!
testNumberLiteralMutation5
	
	self
		mutation: MENumberLiteralMutation
		method: MEMutationTestObject class -> #returnNumberThree
		nonMutated: [ self assert: 222 equals: MEMutationTestObject returnNumberThree ]
		parameters: {
			#nodeSelection -> 2.
			#replacementMode -> #increment.
			#nodeSelectionByIndex -> true.
			#modificationValue -> 123.
		}
		do: [ self assert: 345 equals: MEMutationTestObject returnNumberThree ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:33.35'!
testNumberLiteralMutation6
	
	self
		mutation: MENumberLiteralMutation
		method: MEMutationTestObject class -> #returnNumberThree
		nonMutated: [ self assert: 222 equals: MEMutationTestObject returnNumberThree ]
		parameters: {
			#nodeSelection -> 2.
			#replacementMode -> #decrement.
			#nodeSelectionByIndex -> true.
			#modificationValue -> 123.
		}
		do: [ self assert: 99 equals: MEMutationTestObject returnNumberThree ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:33.34'!
testNumberLiteralMutationRandom
	
	self
		randomMutation: MENumberLiteralMutation
		method: MEMutationTestObject class -> #returnNumberOne
		nonMutated: [ self assert: 222 equals: MEMutationTestObject returnNumberOne ]
		do: [ self assert: 222 ~= MEMutationTestObject returnNumberOne ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:39.57'!
testStringLiteralMutation1
	
	self
		mutation: MEStringLiteralMutation
		method: MEMutationTestObject class -> #returnStringThree
		nonMutated: [ self assert: 'a-string' equals: MEMutationTestObject returnStringThree ]
		parameters: {
			#nodeSelection -> 2.
			#replacementMode -> 'mutatedString'
		}
		do: [ self assert: 'mutatedString' equals: MEMutationTestObject returnStringThree ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:39.603'!
testStringLiteralMutation2
	
	self
		mutation: MEStringLiteralMutation
		method: MEMutationTestObject class -> #returnStringThree
		nonMutated: [ self assert: 'a-string' equals: MEMutationTestObject returnStringThree ]
		parameters: {
			#nodeSelection -> 2.
			#replacementMode -> #randomlyReplaceCharacters
		}
		do: [
			self deny: 'a-string' = MEMutationTestObject returnStringThree.
			self assert: 'a-string' size equals: MEMutationTestObject returnStringThree size ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:39.559'!
testStringLiteralMutation3
	
	self
		mutation: MEStringLiteralMutation
		method: MEMutationTestObject class -> #returnStringThree
		nonMutated: [ self assert: 'a-string' equals: MEMutationTestObject returnStringThree ]
		parameters: {
			#nodeSelection -> 2.
			#replacementMode -> #shuffle
		}
		do: [
			self deny: 'a-string' = MEMutationTestObject returnStringThree.
			self assert: 'a-string' asSortedCollection equals: MEMutationTestObject returnStringThree asSortedCollection ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:39.58'!
testStringLiteralMutation4
	
	self
		mutation: MEStringLiteralMutation
		method: MEMutationTestObject class -> #returnStringThree
		nonMutated: [ self assert: 'a-string' equals: MEMutationTestObject returnStringThree ]
		parameters: {
			#nodeSelection -> 2.
			#replacementMode -> #randomString
		}
		do: [ | length |
			self deny: 'a-string' = MEMutationTestObject returnStringThree.
			length := MEMutationTestObject returnStringThree size.
			self assert: length >= 1 & (length <= ('a-string' size * 2)) ]! !

!MEMutationTests methodsFor: 'tests - TokenMutations' stamp: 'ag 7/21/2012 01:28:39.549'!
testStringLiteralMutationRandom
	
	self
		randomMutation: MEStringLiteralMutation
		method: MEMutationTestObject class -> #returnStringOne
		nonMutated: [ self assert: 'a-string' equals: MEMutationTestObject returnStringOne ]
		do: [ self deny: MEMutationTestObject returnStringOne = 'a-string' ]! !
MEMutationTestSubObject initialize!
